"""Tests for diversity.similarity"""

from collections import defaultdict

from numpy import (
    allclose,
    ndarray,
    array,
    dtype,
    memmap,
    inf,
    float32,
    zeros,
    identity,
    maximum,
    exp,
    random,
    empty,
    empty_like,
)
from numpy.linalg import norm
from pandas import DataFrame, read_csv
import scipy.sparse  # type: ignore[import]
from pytest import fixture, raises, mark

from greylock.log import LOGGER
from greylock.similarity import (
    SimilarityIdentity,
    SimilarityFromArray,
    SimilarityFromDataFrame,
    SimilarityFromFile,
    IntersetSimilarityFromFile,
    SimilarityFromFunction,
    SimilarityFromSymmetricFunction,
    IntersetSimilarityFromFunction,
    weighted_similarity_chunk_nonsymmetric,
    weighted_similarity_chunk_symmetric,
)
from greylock import Metacommunity
from greylock.exceptions import InvalidArgumentError
from greylock.abundance import make_abundance


@fixture
def similarity_function():
    return lambda a, b: 1 / sum(a * b)


rng = random.default_rng()


def similarity_from_distance(a, b):
    return exp(-1 * norm(a - b))


similarity_array_3by3_1 = array(
    [
        [1, 0.5, 0.1],
        [0.5, 1, 0.2],
        [0.1, 0.2, 1],
    ]
)
similarity_array_3by3_2 = array(
    [
        [1.0, 0.9, 0.1],
        [0.9, 1.0, 0.2],
        [0.1, 0.2, 1.0],
    ]
)
similarity_dataframe_3by3 = DataFrame(
    data=similarity_array_3by3_1,
    columns=["species_1", "species_2", "species_3"],
    index=["species_1", "species_2", "species_3"],
)
similarity_filecontent_3by3_tsv = (
    "species_1\tspecies_2\tspecies_3\n"
    "1.0\t0.5\t0.1\n"
    "0.5\t1.0\t0.2\n"
    "0.1\t0.2\t1.0\n"
)
similarities_filecontents_3by3_csv = (
    "species_1,species_2,species_3\n" "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
)
# fmt: off
similarities_filecontents_32by3_csv = (
    "species_1,species_2,species_3\n"
    "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
    "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
    "1.0,0.5,0.1\n" "0.1,0.2,1.0\n"
    "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
    "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
    "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
    "1.0,0.5,0.1\n" "0.5,1.0,0.2\n" "0.1,0.2,1.0\n"
    "0.0,0.5,0.0\n" "0.5,0.0,0.2\n" "0.0,0.2,0.0\n"
    "0.0,0.5,0.0\n" "0.5,0.0,0.2\n" "0.0,0.2,0.0\n"
    "0.0,0.5,0.0\n" "0.5,0.0,0.2\n" "0.0,0.2,0.0\n"
    "0.0,0.5,0.0\n" "0.5,0.0,0.2\n" "0.0,0.2,0.0\n"
)
# fmt: on
similarity_sparse_entries = {
    "shape": (3, 3),
    "row": array(
        [
            0,
            1,
            2,
            0,
            1,
            0,
            2,
        ]
    ),
    "col": array(
        [
            0,
            1,
            2,
            1,
            0,
            2,
            0,
        ]
    ),
    "data": array(
        [
            1,
            1,
            1,
            0.5,
            0.5,
            0.3,
            0.3,
        ]
    ),
}
relative_abundance_3by1 = array([[1 / 1000], [1 / 10], [10]])
relative_abundance_3by2 = array([[1 / 1000, 1 / 100], [1 / 10, 1 / 1], [10, 100]])
relative_abundance_3by2_2 = array(
    [
        [0.7, 0.3],
        [0.1, 0.3],
        [0.2, 0.4],
    ]
)
weighted_abundances_3by1_1 = array([[1.051], [2.1005], [10.0201]])
weighted_abundances_3by1_2 = array([[0.35271989], [0.13459705], [0.0601738]])
weighted_abundances_3by1_3 = array([[0.35271989], [0.13459705], [0.0601738]])
weighted_abundances_3by2_1 = array(
    [[1.051, 10.51], [2.1005, 21.005], [10.0201, 100.201]]
)
weighted_abundances_3by2_2 = array(
    [
        [0.81, 0.61],
        [0.77, 0.65],
        [0.29, 0.49],
    ]
)
weighted_abundances_3by2_3 = (
    array(
        [
            [0.35271989, 3.52719894],
            [0.13459705, 1.34597047],
            [0.0601738, 0.60173802],
        ]
    ),
)
weighted_abundances_3by2_4 = array(
    [
        [1.46290476, 14.62904762],
        [0.48763492, 4.87634921],
        [0.20898639, 2.08986395],
    ]
)
X_3by1 = array([[1], [3], [7]])
X_3by2 = array([[1, 2], [3, 5], [7, 11]])


@fixture
def memmapped_similarity_matrix(tmp_path):
    memmapped = memmap(
        tmp_path / "similarity_matrix.npy",
        dtype=dtype("f8"),
        mode="w+",
        offset=0,
        shape=similarity_array_3by3_1.shape,
        order="C",
    )
    memmapped[:, :] = similarity_array_3by3_1
    return memmapped


@fixture
def make_similarity_from_file(tmp_path):
    def make(
        filename="similarity_matrix.tsv",
        filecontent=similarity_filecontent_3by3_tsv,
        chunk_size=1,
        similarity_class=SimilarityFromFile,
    ):
        filepath = tmp_path / filename
        with open(filepath, "w") as file:
            file.write(filecontent)
        return similarity_class(similarity_file_path=filepath, chunk_size=chunk_size)

    return make


@mark.parametrize(
    "relative_abundance, expected, kwargs",
    [
        (relative_abundance_3by2, weighted_abundances_3by2_1, {}),
        (relative_abundance_3by2, weighted_abundances_3by2_1, {"chunk_size": 2}),
        (
            relative_abundance_3by2,
            weighted_abundances_3by2_1,
            {
                "filename": "similarity_matrix.csv",
                "filecontent": similarities_filecontents_3by3_csv,
            },
        ),
        (relative_abundance_3by1, weighted_abundances_3by1_1, {}),
    ],
)
def test_weighted_abundances(
    relative_abundance, expected, kwargs, make_similarity_from_file
):
    similarity = make_similarity_from_file(**kwargs)
    if similarity.path.name.endswith("tsv"):
        sep = "\t"
    else:
        sep = ","
    expected_similarity_matrix = read_csv(similarity.path, sep=sep).to_numpy()
    similarities_out = empty_like(expected_similarity_matrix)
    similarity.similarities_out = similarities_out
    actual = similarity.weighted_abundances(relative_abundance)
    assert allclose(expected_similarity_matrix, similarities_out)
    assert allclose(actual, expected)


def test_nonsquare_from_file(make_similarity_from_file):
    sim = make_similarity_from_file(similarity_class=IntersetSimilarityFromFile)
    counts = array([[1], [1], [1]])
    with raises(InvalidArgumentError):
        Metacommunity(counts, sim).to_dataframe(viewpoint=0)


def test_interset_from_file(make_similarity_from_file):
    sim = make_similarity_from_file(
        filecontent=similarities_filecontents_32by3_csv,
        chunk_size=4,
        similarity_class=IntersetSimilarityFromFile,
    )
    counts = array([[50], [25], [25]])
    abundance_object = make_abundance(counts, False)
    # fmt: off
    expected = array([[0.65, 0.55, 0.35,
                      0.65, 0.55, 0.35,
                      0.65, 0.35,
                      0.65, 0.55, 0.35,
                      0.65, 0.55, 0.35,
                      0.65, 0.55, 0.35,
                      0.65, 0.55, 0.35,
                      0.125, 0.3, 0.05,
                      0.125, 0.3, 0.05,
                      0.125, 0.3, 0.05,
                       0.125, 0.3, 0.05]]).T
    # fmt: on
    result = sim @ abundance_object
    assert allclose(result, expected)
    expected_similarity_matrix = read_csv(sim.path).to_numpy()
    similarities_out = empty_like(expected_similarity_matrix)
    sim.similarities_out = similarities_out
    result = sim.weighted_abundances(
        abundance_object.normalized_subcommunity_abundance,
    )
    assert allclose(result, expected)
    assert allclose(expected_similarity_matrix, similarities_out)


@mark.parametrize(
    "similarity, simclass, relative_abundance, expected",
    [
        (
            similarity_dataframe_3by3,
            SimilarityFromDataFrame,
            relative_abundance_3by2,
            weighted_abundances_3by2_1,
        ),
        (
            similarity_dataframe_3by3,
            SimilarityFromDataFrame,
            relative_abundance_3by1,
            weighted_abundances_3by1_1,
        ),
        (
            similarity_array_3by3_1,
            SimilarityFromArray,
            relative_abundance_3by1,
            weighted_abundances_3by1_1,
        ),
        (
            similarity_array_3by3_2,
            SimilarityFromArray,
            relative_abundance_3by2_2,
            weighted_abundances_3by2_2,
        ),
    ],
)
def test_weighted_abundances_from_array(
    similarity, simclass, relative_abundance, expected
):
    similarity = simclass(similarity=similarity)
    weighted_abundances = similarity.weighted_abundances(
        relative_abundance=relative_abundance
    )
    assert allclose(weighted_abundances, expected)


def test_weighted_abundances_from_memmap(memmapped_similarity_matrix):
    similarity = SimilarityFromArray(similarity=memmapped_similarity_matrix)
    weighted_abundances = similarity.weighted_abundances(
        relative_abundance=relative_abundance_3by2
    )
    assert allclose(weighted_abundances, weighted_abundances_3by2_1)


@mark.parametrize(
    "relative_abundance, X, chunk_size, expected",
    [
        (relative_abundance_3by2, X_3by2, 2, weighted_abundances_3by2_3),
        (relative_abundance_3by2, X_3by1, 1, weighted_abundances_3by2_4),
        (relative_abundance_3by1, X_3by2, 4, weighted_abundances_3by1_2),
        (relative_abundance_3by1, X_3by2, 2, weighted_abundances_3by1_2),
    ],
)
def test_weighted_abundances_from_function(
    relative_abundance, similarity_function, X, chunk_size, expected
):
    species_count = X.shape[0]
    similarities_out = empty((species_count, species_count))
    expected_similarities = array(
        [
            [similarity_function(X[i], X[j]) for i in range(species_count)]
            for j in range(species_count)
        ]
    )
    similarity = SimilarityFromFunction(
        func=similarity_function,
        X=X,
        chunk_size=chunk_size,
        similarities_out=similarities_out,
    )
    weighted_abundances = similarity.weighted_abundances(
        relative_abundance=relative_abundance
    )
    assert allclose(weighted_abundances, expected)
    assert allclose(similarities_out, expected_similarities)


def test_weighted_similarity_chunk(similarity_function):
    chunk_index, chunk, similarities = weighted_similarity_chunk_nonsymmetric(
        similarity=similarity_function,
        X=X_3by2,
        Y=None,
        relative_abundance=relative_abundance_3by2,
        chunk_size=3,
        chunk_index=0,
    )
    assert chunk_index == 0
    assert allclose(chunk, weighted_abundances_3by2_3)


def make_array(spec, array_class=zeros):
    a = array_class(spec["shape"], dtype=float32)
    for i, value in enumerate(spec["data"]):
        a[spec["row"][i], spec["col"][i]] = value
    return a


def compare_dense_sparse(counts, dense_similarity, sparse_similarity):
    viewpoints = [0, 1, 2, inf]
    measures = (
        "alpha",
        "rho",
        "beta",
        "gamma",
        "normalized_alpha",
        "normalized_rho",
        "normalized_beta",
    )
    meta_dense = Metacommunity(counts, similarity=dense_similarity)
    meta_dense_df = meta_dense.to_dataframe(viewpoint=viewpoints, measures=measures)
    meta_sparse = Metacommunity(
        counts, similarity=SimilarityFromArray(sparse_similarity)
    )
    meta_sparse_df = meta_sparse.to_dataframe(viewpoint=viewpoints, measures=measures)
    for col in meta_dense_df:
        if col != "community":
            assert allclose(meta_dense_df[col], meta_sparse_df[col])


@mark.parametrize(
    "sparse_class",
    [
        scipy.sparse.bsr_array,
        scipy.sparse.coo_array,
        scipy.sparse.csc_array,
        scipy.sparse.csr_array,
        scipy.sparse.bsr_matrix,
        scipy.sparse.coo_matrix,
        scipy.sparse.csc_matrix,
        scipy.sparse.csr_matrix,
    ],
)
def test_sparse_similarity(sparse_class):
    spec = similarity_sparse_entries
    dense_similarity = make_array(spec)
    sparse_similarity = sparse_class(
        (spec["data"], (spec["row"], spec["col"])), shape=spec["shape"]
    )
    counts = DataFrame({"Medford": [3, 2, 0], "Somerville": [1, 4, 0]})
    compare_dense_sparse(counts, dense_similarity, sparse_similarity)


@mark.parametrize("sparse_class", [scipy.sparse.dia_array, scipy.sparse.dia_matrix])
def test_diag_sparse(sparse_class):
    data = array([[0.5] * 4, [1] * 4, [0.5] * 4])
    offsets = array([-1, 0, 1])
    dense_similarity = array(
        [
            [1.0, 0.5, 0.0, 0.0],
            [0.5, 1.0, 0.5, 0.0],
            [0.0, 0.5, 1.0, 0.5],
            [0.0, 0.0, 0.5, 1.0],
        ]
    )
    sparse_similarity = sparse_class((data, offsets), shape=(4, 4))
    counts = DataFrame({"Cambridge": [5, 2, 0, 9], "Boston": [2, 3, 3, 2]})
    compare_dense_sparse(counts, dense_similarity, sparse_similarity)


@mark.parametrize(
    "sparse_class",
    [
        scipy.sparse.lil_array,
        scipy.sparse.lil_matrix,
        scipy.sparse.dok_array,
        scipy.sparse.dok_matrix,
    ],
)
def test_incremental_sparse(sparse_class):
    spec = similarity_sparse_entries
    dense_similarity = make_array(spec)
    sparse_similarity = make_array(spec, sparse_class)
    counts = DataFrame({"Arlington": [23, 12, 8], "Watertown": [15, 14, 19]})
    compare_dense_sparse(counts, dense_similarity, sparse_similarity)


def another_similarity_func(row_i, row_j):
    i = row_i[0] / row_i[1]
    j = row_j[0] / row_j[1]
    return min(1.0, max(0.0, 0.4 * (3 - abs(i - j))))


symmetric_example_X = array([[0, 88], [10, 10], [200, 100], [99, 33]])
symmetric_example_abundance = array([[1, 0], [0, 1], [1, 0], [0, 10]])


@mark.parametrize(
    "chunk_index, expected",
    [
        (0, [[0.4, 0.8], [1.6, 4.0], [0.4, 0.8], [0.0, 0.4]]),
        (2, [[0.0, 0.0], [0.0, 0.0], [0.0, 8.0], [0.8, 0.0]]),
    ],
)
def test_weighted_similarity_chunk_symmetric(chunk_index, expected):
    _, result, _ = weighted_similarity_chunk_symmetric(
        another_similarity_func,
        symmetric_example_X,
        symmetric_example_abundance,
        2,
        chunk_index,
    )
    assert allclose(result, array(expected))


def test_symmetric_similarity():
    species_count = symmetric_example_X.shape[0]
    similarities_out = empty((species_count, species_count))
    expected = array([[1.4, 0.8], [1.6, 5.0], [1.4, 8.8], [0.8, 10.4]])
    obj = SimilarityFromSymmetricFunction(
        func=another_similarity_func,
        X=symmetric_example_X,
        chunk_size=2,
        similarities_out=similarities_out,
    )
    expected_similarities = array(
        [
            [
                another_similarity_func(symmetric_example_X[i], symmetric_example_X[j])
                for i in range(species_count)
            ]
            for j in range(species_count)
        ]
    )
    result = obj.weighted_abundances(
        symmetric_example_abundance,
    )
    assert allclose(result, expected)
    assert allclose(similarities_out, expected_similarities)


animal_features = DataFrame(
    {
        "breathes": [
            "air",
            "air",
            "air",
            "air",
            "water",
            "air",
            "water",
            "air",
            "air",
            "air",
            "air",
            "air",
        ],
        "covering": [
            "fur",
            "fur",
            "fur",
            "scales",
            "scales",
            "fur",
            "scales",
            "feathers",
            "fur",
            "fur",
            "scales",
            "fur",
        ],
        "diet": [
            "meat",
            "omni",
            "plants",
            "meat",
            "omni",
            "plants",
            "omni",
            "omni",
            "meat",
            "plants",
            "plants",
            "plants",
        ],
        "n_legs": [
            4,
            4,
            4,
            0,
            0,
            4,
            0,
            2,
            4,
            4,
            4,
            2,
        ],
    },
    index=[
        "cat",
        "dog",
        "rabbit",
        "snake",
        "goldfish",
        "gerbil",
        "betafish",
        "chicken",
        "tiger",
        "giraffe",
        "turtle",
        "monkey",
    ],
)

animal_communities = DataFrame(
    {
        "zoo": [
            0,
            0,
            2,
            6,
            0,
            0,
            0,
            0,
            2,
            2,
            4,
            5,
        ],
        "petco": [
            4,
            6,
            4,
            4,
            40,
            8,
            30,
            0,
            0,
            0,
            1,
            0,
        ],
        "shelter": [
            10,
            15,
            1,
            1,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
        ],
    },
    index=[
        "cat",
        "dog",
        "rabbit",
        "snake",
        "goldfish",
        "gerbil",
        "betafish",
        "chicken",
        "tiger",
        "giraffe",
        "turtle",
        "monkey",
    ],
)


def feature_similarity(animal_i, animal_j):
    if animal_i.breathes != animal_j.breathes:
        return 0.0
    if animal_i.covering == animal_j.covering:
        if animal_i.Index == animal_j.Index:
            result = 1
        else:
            result = 0.98
    else:
        result = 0.5
    if animal_i.n_legs != animal_j.n_legs:
        result *= 0.5
    if animal_i.diet != animal_j.diet:
        if "omni" in (animal_i.diet, animal_j.diet):
            result *= 0.9
        else:
            result *= 0.7
    return result


def animal_similarity_matrix():
    columns = defaultdict(list)
    index = []
    results = []
    for i, animal_i in enumerate(animal_features.itertuples()):
        index.append(animal_i.Index)
        for j, animal_j in enumerate(animal_features.itertuples()):
            s = feature_similarity(animal_i, animal_j)
            columns[animal_j.Index].append(s)
            results.append((s, animal_i.Index, animal_j.Index))
    return DataFrame(columns, index=index)


def test_feature_similarity():
    measures = [
        "alpha",
        "rho",
        "beta",
        "gamma",
        "normalized_alpha",
        "normalized_rho",
        "normalized_beta",
        "rho_hat",
    ]
    viewpoints = [0, 1, 2, inf]
    m = Metacommunity(
        animal_communities,
        similarity=SimilarityFromDataFrame(animal_similarity_matrix()),
    )
    df1 = m.to_dataframe(viewpoint=viewpoints, measures=measures).set_index(
        ["community", "viewpoint"]
    )
    m = Metacommunity(
        animal_communities,
        SimilarityFromFunction(
            func=feature_similarity,
            X=animal_features,
            chunk_size=4,
        ),
    )
    df2 = m.to_dataframe(viewpoint=viewpoints, measures=measures).set_index(
        ["community", "viewpoint"]
    )
    assert allclose(df1.to_numpy(), df2.to_numpy())
    m = Metacommunity(
        animal_communities,
        SimilarityFromSymmetricFunction(
            func=feature_similarity,
            X=animal_features,
            chunk_size=4,
        ),
    )
    df3 = m.to_dataframe(viewpoint=viewpoints, measures=measures).set_index(
        ["community", "viewpoint"]
    )
    assert allclose(df1.to_numpy(), df3.to_numpy())


def test_nonsymmetric():
    """
    SimilarityFromFunction will yield, if the function is not
    actually symmetric, a matrix that is not a reflection of itself
    across the diagonal.
    """

    def nonsym_similarity_function(species_i, species_j):
        diff = maximum((species_i - species_j), 0)
        return 1 / (1 + norm(diff) / 100)

    X = array(
        [[54, 200, 45, 123], [55, 67, 44, 99], [25, 145, 56, 12], [154, 98, 55, 98]]
    )
    counts = identity(4)
    sim = SimilarityFromFunction(nonsym_similarity_function, X)
    matrix = zeros(shape=(4, 4))
    for i in range(4):
        for j in range(4):
            matrix[i, j] = nonsym_similarity_function(X[i], X[j])
    result = sim.weighted_abundances(counts)
    assert allclose(result, matrix)
    for i in range(4):
        for j in range(i + 1, 4):
            assert result[i, j] != result[j, i]


@fixture
def callcounter():
    return defaultdict(int)


@mark.parametrize("n", [1, 3, 10, 37])
def test_identity_similarity(n):
    abundance = rng.random((n, 8))
    similarities_out = empty(shape=(n, n))
    sim = SimilarityIdentity(similarities_out=similarities_out)
    result = sim.weighted_abundances(abundance)
    assert (result == abundance).all()
    assert (similarities_out == identity(n)).all()


@mark.parametrize("a", [similarity_array_3by3_1, similarity_array_3by3_2])
def test_array_similarity(a):
    abundance = rng.random((a.shape[0], 10))
    similarities_out = empty_like(a)
    sim = SimilarityFromArray(a, similarities_out=similarities_out)
    sim.weighted_abundances(abundance)
    assert (similarities_out == a).all()


def test_dataframe_similarity():
    labels_2a = [
        "owl",
        "eagle",
        "flamingo",
        "swan",
        "duck",
        "chicken",
        "turkey",
        "dodo",
        "dove",
    ]
    no_species_2a = len(labels_2a)
    S_2a = identity(n=no_species_2a)
    S_2a[0][1:9] = (0.91, 0.88, 0.88, 0.88, 0.88, 0.88, 0.88, 0.88)  # owl
    S_2a[1][2:9] = (0.88, 0.89, 0.88, 0.88, 0.88, 0.89, 0.88)  # eagle
    S_2a[2][3:9] = (0.90, 0.89, 0.88, 0.88, 0.88, 0.89)  # flamingo
    S_2a[3][4:9] = (0.92, 0.90, 0.89, 0.88, 0.88)  # swan
    S_2a[4][5:9] = (0.91, 0.89, 0.88, 0.88)  # duck
    S_2a[5][6:9] = (0.92, 0.88, 0.88)  # chicken
    S_2a[6][7:9] = (0.89, 0.88)  # turkey
    S_2a[7][8:9] = 0.88  # dodo
    S_2a = maximum(S_2a, S_2a.transpose())
    S_2a_df = DataFrame(
        {labels_2a[i]: S_2a[i] for i in range(no_species_2a)}, index=labels_2a
    )
    abundance = rng.random((no_species_2a, 10))
    similarities_out = empty_like(S_2a)
    sim = SimilarityFromDataFrame(S_2a_df, similarities_out=similarities_out)
    sim.weighted_abundances(abundance)
    assert (similarities_out == S_2a).all()


@mark.parametrize(
    "sim, key, expected_count",
    [
        [SimilarityIdentity(), "identity", 3],
        ["array", "array", 3],
        ["df", "df", 3],
        ["file", "file", 1],
        [
            SimilarityFromFunction(
                similarity_from_distance, X=array([[1, 2], [2, 1], [3, 0]])
            ),
            "func",
            1,
        ],
        [
            SimilarityFromSymmetricFunction(
                similarity_from_distance, X=array([[1, 2], [2, 1], [3, 0]])
            ),
            "sfunc",
            1,
        ],
    ],
)
def test_compuation_count(
    sim, key, expected_count, callcounter, make_similarity_from_file
):
    """
    Test that we unify the abundance array and call weighted_abundances only once
    for the subclasses of Similarity that are expensive to run, but don't bother
    with that for trivial subclasses.
    """

    def count_decorator(f, counter, key):
        def wrapper(*args, **kwds):
            counter[key] += 1
            return f(*args, **kwds)

        return wrapper

    abundances = array([[1, 2], [2, 5], [9, 3]])
    if sim == "file":
        sim = make_similarity_from_file()
    elif sim == "array":
        sim = SimilarityFromArray(similarity_array_3by3_1)
    elif sim == "df":
        sim = SimilarityFromDataFrame(similarity_dataframe_3by3)

    sim.weighted_abundances = count_decorator(sim.weighted_abundances, callcounter, key)
    m = Metacommunity(abundances, sim)
    m.metacommunity_diversity(viewpoint=1, measure="alpha")
    m.to_dataframe(viewpoint=0)

    assert callcounter[key] == expected_count


@mark.parametrize(
    "X, Y, abundance, expected",
    [
        [
            array([[2, 1, 5], [4, 3, 2]]),
            array([[2, 0, 5], [4, 2, 1], [2, 0, 5]]),
            array([[0.5, 0.1], [0.25, 0.1], [0.25, 0.8]]),
            array([[0.27846671, 0.33211435], [0.06766633, 0.03257625]]),
        ],
        [
            array([[1, 0, 0]]),
            array([[1, 0, 0], [0, 1, 0], [0, 1, 1]]),
            array([[1.0, 0.8, 0.0], [0.0, 0.1, 0.5], [0.0, 0.1, 0.5]]),
            array([[1.0, 0.84200379, 0.21001897]]),
        ],
    ],
)
def test_interset_similarity(X, Y, abundance, expected):
    sim = IntersetSimilarityFromFunction(similarity_from_distance, X, Y)
    result = sim.weighted_abundances(abundance)
    assert allclose(result, expected)
    abundance_obj = make_abundance(abundance, False)
    result = sim @ abundance_obj
    assert allclose(result, expected)


def test_square_similarity_requirement():
    sim = SimilarityFromArray(array([[1.0, 0.4, 0.6], [0.2, 1.0, 0.9]]))
    counts = array([[1], [1], [1]])
    with raises(InvalidArgumentError):
        Metacommunity(counts, sim).to_dataframe(viewpoint=0)


def test_interset_diversity_forbidden():
    sim = IntersetSimilarityFromFunction(
        similarity_from_distance,
        X=array([[1, 0, 0]]),
        Y=array([[0, 1, 0], [0, 1, 0], [0, 1, 0], [0, 1, 0], [0, 1, 0]]),
    )
    counts = array([[1, 1, 1, 1, 1]])
    with raises(InvalidArgumentError):
        Metacommunity(counts, sim).to_dataframe(viewpoint=0)


def test_matmul():
    similarity_df = DataFrame(
        {"bee": [0.6, 0.22], "butterfly": [0.55, 0.27], "lobster": [0.45, 0.28]},
        index=["ladybug", "fish"],
    )
    similarity = SimilarityFromDataFrame(similarity_df)
    abundance = make_abundance(array([[5000], [2000], [3000]]), False)
    expected = array([[0.545], [0.248]])
    relative_abundance = similarity @ abundance
    assert allclose(expected, relative_abundance)



"""
Testing the testing infrastructure!
"""

from greylock.tests.mockray import put, remote, wait, get


def test_mock():
    @remote
    def interesting(n):
        return [(n**k) for k in range(5)]

    thing = (5, "foobar")
    assert put(thing) == thing

    future = interesting.remote(3)
    one_result = get(future)
    assert len(one_result) == 5

    futures = []
    results = []
    for n in range(10):
        if len(futures) > 4:
            ready_refs, futures = wait(futures)
            results += get(ready_refs)
        futures.append(interesting.remote(n))
    results += get(futures)
    assert len(results) == 10
    results.sort()
    for n, result in enumerate(results):
        assert result[1] == n



from numpy import (
    sum,
    sqrt,
    allclose,
    ndarray,
    array,
    dtype,
    memmap,
    inf,
    float32,
    zeros,
    empty,
    dot,
)
from numpy.linalg import norm
import ray
from greylock.similarity import (
    SimilarityFromArray,
    SimilarityFromFunction,
    SimilarityFromSymmetricFunction,
)
from greylock.ray import (
    SimilarityFromRayFunction,
    SimilarityFromSymmetricRayFunction,
    IntersetSimilarityFromRayFunction,
)
import greylock.tests.mockray as mockray
from greylock.tests.base_tests.similarity_test import similarity_from_distance
from pytest import fixture, raises, mark
from greylock.tests.base_tests.similarity_test import (
    relative_abundance_3by2,
    relative_abundance_3by1,
    X_3by1,
    X_3by2,
)
from greylock import Metacommunity
from greylock.exceptions import InvalidArgumentError


def ray_fix(monkeypatch):
    monkeypatch.setattr(ray, "put", mockray.put)
    monkeypatch.setattr(ray, "get", mockray.get)
    monkeypatch.setattr(ray, "remote", mockray.remote)
    monkeypatch.setattr(ray, "wait", mockray.wait)


@fixture(autouse=True)
def setup(monkeypatch):
    ray_fix(monkeypatch)


MEASURES = (
    "alpha",
    "rho",
    "beta",
    "gamma",
    "normalized_alpha",
    "normalized_rho",
    "normalized_beta",
    "rho_hat",
)

abundances_large = array(
    [
        [45, 23],
        [4, 54],
        [23, 1],
        [623, 0],
        [23, 7],
        [23, 90],
        [1, 1],
        [34, 62],
        [13, 72],
        [23, 23],
        [72, 3],
        [62, 3],
        [623, 4],
        [234, 90],
        [23, 12],
        [96, 5],
        [6, 24],
        [6, 4],
        [65, 91],
        [345, 4],
        [23, 62],
        [62, 73],
        [23, 7],
        [23, 90],
        [1, 1],
        [34, 62],
        [13, 72],
        [23, 23],
        [72, 3],
        [62, 3],
        [623, 4],
        [234, 90],
        [23, 12],
        [13, 72],
        [23, 23],
        [72, 3],
        [62, 3],
        [623, 4],
        [234, 90],
        [23, 12],
        [96, 5],
        [6, 24],
        [6, 4],
        [65, 91],
        [345, 4],
        [23, 62],
        [62, 73],
        [23, 7],
        [23, 90],
        [1, 1],
        [34, 62],
        [13, 72],
        [23, 23],
        [72, 3],
        [62, 3],
        [623, 4],
        [234, 90],
        [23, 12],
        [96, 5],
        [6, 24],
        [6, 4],
        [65, 91],
        [345, 4],
        [23, 62],
        [62, 73],
        [23, 7],
    ]
)
X_large = array(
    [
        [6, 1, 1],
        [7, 34, 62],
        [8, 13, 72],
        [9, 23, 23],
        [11, 72, 3],
        [12, 62, 3],
        [13, 623, 4],
        [14, 234, 90],
        [34, 62, 3],
        [62, 43, 4],
        [23, 34, 90],
        [1, 23, 12],
        [2, 96, 5],
        [0, 45, 23],
        [1, 4, 54],
        [2, 23, 1],
        [3, 623, 0],
        [4, 23, 7],
        [5, 23, 90],
        [3, 6, 24],
        [5, 6, 4],
        [6, 65, 91],
        [34, 75, 4],
        [1, 23, 62],
        [2, 62, 73],
        [3, 23, 7],
        [15, 23, 12],
        [16, 96, 5],
        [17, 6, 24],
        [18, 6, 4],
        [19, 65, 91],
        [3, 45, 4],
        [20, 23, 62],
        [21, 62, 73],
        [22, 23, 7],
        [23, 84, 90],
        [14, 1, 1],
        [24, 34, 62],
        [25, 13, 72],
        [23, 75, 23],
        [26, 72, 3],
        [27, 62, 3],
        [62, 73, 4],
        [21, 34, 90],
        [21, 23, 12],
        [22, 13, 72],
        [45, 23, 23],
        [23, 72, 3],
        [24, 62, 3],
        [62, 63, 4],
        [24, 34, 90],
        [24, 23, 12],
        [45, 96, 5],
        [24, 6, 24],
        [25, 6, 4],
        [26, 65, 91],
        [35, 45, 4],
        [23, 23, 62],
        [27, 62, 73],
        [28, 23, 7],
        [29, 23, 90],
        [4, 1, 1],
        [29, 34, 62],
        [31, 13, 72],
        [32, 3, 23],
        [33, 2, 3],
    ]
)


def similarity_function(a, b):
    a = a / norm(a)
    b = b / norm(b)
    return dot(a, b)


@mark.parametrize(
    "x, y, expected",
    [
        (array([1, 2, 3]), array([1, 2, 3]), 1.0),
        (array([0, 1, 0]), array([2, 0, 0]), 0.0),
        (array([0, 1, 1]), array([0, 1, 0]), 1 / sqrt(2)),
        (array([0, 1, 1]), array([1, 1, 0]), 1 / 2),
    ],
)
def test_similarity_function(x, y, expected):
    assert allclose(similarity_function(x, y), expected)


@mark.parametrize(
    "relative_abundance, X, chunk_size",
    [
        (relative_abundance_3by2, X_3by2, 2),
        (relative_abundance_3by2, X_3by1, 1),
        (relative_abundance_3by1, X_3by2, 4),
        (relative_abundance_3by1, X_3by2, 2),
    ],
)
def test_weighted_abundances_from_function(relative_abundance, X, chunk_size):
    sim_matrix = zeros(shape=(X.shape[0], X.shape[0]))
    for i in range(X.shape[0]):
        for j in range(X.shape[0]):
            sim_matrix[i, j] = similarity_function(X[i], X[j])
    similarity1 = SimilarityFromArray(sim_matrix)
    expected = similarity1.weighted_abundances(relative_abundance=relative_abundance)
    similarities_out = empty(shape=(X.shape[0], X.shape[0]))
    similarity = SimilarityFromRayFunction(
        func=similarity_function,
        X=X,
        chunk_size=chunk_size,
        similarities_out=similarities_out,
    )
    weighted_abundances = similarity.weighted_abundances(
        relative_abundance=relative_abundance
    )
    assert allclose(weighted_abundances, expected)
    assert allclose(similarities_out, sim_matrix)


def test_comparisons():
    results = []
    for simclass in [
        SimilarityFromFunction,
        SimilarityFromRayFunction,
        SimilarityFromSymmetricFunction,
        SimilarityFromSymmetricRayFunction,
    ]:
        if "Ray" in simclass.__name__:
            similarity = simclass(
                func=similarity_function, X=X_large, chunk_size=4, max_inflight_tasks=2
            )
        else:
            similarity = simclass(func=similarity_function, X=X_large, chunk_size=4)
        m = Metacommunity(abundances_large, similarity)
        df = m.to_dataframe(viewpoint=[0, 1, 2, 200], measures=MEASURES)
        results.append(df.drop(columns="community"))
    for result in results[1:]:
        assert allclose(results[0].to_numpy(), result.to_numpy())


def test_similarities_out():
    computed_similarity_matrices = []
    similarities_out = empty((X_large.shape[0], X_large.shape[0]))
    for simclass in [
        SimilarityFromFunction,
        SimilarityFromRayFunction,
        SimilarityFromSymmetricFunction,
        SimilarityFromSymmetricRayFunction,
    ]:
        if "Ray" in simclass.__name__:
            similarity = simclass(
                func=similarity_function,
                X=X_large,
                chunk_size=7,
                max_inflight_tasks=5,
                similarities_out=similarities_out,
            )
        else:
            similarity = simclass(
                func=similarity_function,
                X=X_large,
                chunk_size=7,
                similarities_out=similarities_out,
            )
        similarity.weighted_abundances(abundances_large)
        computed_similarity_matrices.append(similarities_out)
    for matrix in computed_similarity_matrices[1:]:
        assert allclose(computed_similarity_matrices[0], matrix)


@mark.parametrize(
    "X, Y, abundance, expected",
    [
        [
            array([[2, 1, 5], [4, 3, 2]]),
            array([[2, 0, 5], [4, 2, 1], [2, 0, 5]]),
            array([[0.5, 0.1], [0.25, 0.1], [0.25, 0.8]]),
            array([[0.27846671, 0.33211435], [0.06766633, 0.03257625]]),
        ],
        [
            array([[1, 0, 0]]),
            array([[1, 0, 0], [0, 1, 0], [0, 1, 1]]),
            array([[1.0, 0.8, 0.0], [0.0, 0.1, 0.5], [0.0, 0.1, 0.5]]),
            array([[1.0, 0.84200379, 0.21001897]]),
        ],
    ],
)
def test_interset_similarity(X, Y, abundance, expected):
    sim = IntersetSimilarityFromRayFunction(similarity_from_distance, X, Y)
    result = sim.weighted_abundances(abundance)
    assert allclose(result, expected)


def test_interset_diversity_forbidden():
    sim = IntersetSimilarityFromRayFunction(
        similarity_from_distance,
        X=array([[1, 0, 0]]),
        Y=array([[0, 1, 0], [0, 1, 0], [0, 1, 0], [0, 1, 0], [0, 1, 0]]),
    )
    counts = array([[1, 1, 1, 1, 1]])
    with raises(InvalidArgumentError):
        Metacommunity(counts, sim).to_dataframe(viewpoint=0)



"""Tests for greylock.parameters."""

from pytest import mark, warns

from greylock.parameters import configure_arguments
from greylock.exceptions import ArgumentWarning

CONFIGURE_ARGUMENTS_TEST_CASES = [
    {
        "description": "Warn viewpoint > 100",
        "args": ["-v", "1", "2", "0", "inf", "101"],
        "expect_warning": True,
    },
    {
        "description": "Do not warn valid viewpoint",
        "args": ["-v", "0", "100", "inf"],
        "expect_warning": False,
    },
]


class TestConfigureArguments:
    """Tests parameters.configure_arguments."""

    @mark.parametrize("test_case", CONFIGURE_ARGUMENTS_TEST_CASES)
    def test_configure_arguments(self, test_case):
        parser = configure_arguments()
        if test_case["expect_warning"]:
            with warns(ArgumentWarning):
                parser.parse_args(test_case["args"])
        else:
            parser.parse_args(test_case["args"])



"""Helper module for logging operations.

Constants
---------
LOG_HANDLER: logging.StreamHandler
    Handler used for logging.
LOGGER: logging.Logger
    Multiprocessing-safe logger.
"""

from logging import (
    Formatter,
    StreamHandler,
)
from multiprocessing import get_logger

DATE_FORMAT = "%Y-%m-%dT%H:%M:%S%z"
LOGGING_FORMAT = (
    "%(asctime)s\t(%(processName)s, %(threadName)s)\t%(levelname)s\t%(message)s"
)

LOGGER = get_logger()
formatter = Formatter(fmt=LOGGING_FORMAT, datefmt=DATE_FORMAT)
LOG_HANDLER = StreamHandler()
LOG_HANDLER.setFormatter(formatter)
LOGGER.addHandler(LOG_HANDLER)



"""Module for calculating relative sub- and metacomunity abundances.

Classes
-------
Abundance
    Relative (normalized) species abundances in (meta-/sub-) communities
AbundanceForDiversity
    Species abundances-- normalized over metacommunity, normalized over each subcommunity,
    and totalled across metacommunity-- as is required for diversity calculations

"""

from functools import cached_property
from collections.abc import Iterable
from typing import Union

from numpy import arange, ndarray, concatenate, minimum
from pandas import DataFrame, RangeIndex
from scipy.sparse import issparse  # type: ignore[import]


class Abundance:
    def __init__(
        self, counts: ndarray, subcommunity_names: Iterable[Union[str, int]]
    ) -> None:
        """
        Parameters
        ----------
        counts:
            2-d array with one column per subcommunity, one row per
            species, containing the count of each species in the
            corresponding subcommunities.
        """
        self.subcommunities_names = subcommunity_names
        self.num_subcommunities = counts.shape[1]
        self.min_count = minimum(1 / counts.sum(), 1e-9)

        self.subcommunity_abundance = self.make_subcommunity_abundance(counts=counts)
        self.normalized_subcommunity_abundance = (
            self.make_normalized_subcommunity_abundance()
        )

    def make_subcommunity_abundance(self, counts: ndarray) -> ndarray:
        """Calculates the relative abundances in subcommunities.

        Parameters
        ----------
        counts
            2-d array with one column per subcommunity, one row per
            species, containing the count of each species in the
            corresponding subcommunities.

        Returns
        -------
        A numpy.ndarray of shape (n_species, n_subcommunities), where
        rows correspond to unique species, columns correspond to
        subcommunities and each element is the abundance of the species
        in the subcommunity relative to the total metacommunity size.
        """
        return counts / counts.sum()

    def make_subcommunity_normalizing_constants(self) -> ndarray:
        """Calculates subcommunity normalizing constants.

        Returns
        -------
        A numpy.ndarray of shape (n_subcommunities,), with the fraction
        of each subcommunity's size of the metacommunity.
        """
        return self.subcommunity_abundance.sum(axis=0)

    def make_normalized_subcommunity_abundance(self) -> ndarray:
        """Calculates normalized relative abundances in subcommunities.

        Returns
        -------
        A numpy.ndarray of shape (n_species, n_subcommunities), where
        rows correspond to unique species, columns correspond to
        subcommunities and each element is the abundance of the species
        in the subcommunity relative to the subcommunity size.
        """
        self.subcommunity_normalizing_constants = (
            self.make_subcommunity_normalizing_constants()
        )
        return self.subcommunity_abundance / self.subcommunity_normalizing_constants

    def premultiply_by(self, similarity):
        return similarity.weighted_abundances(self.normalized_subcommunity_abundance)


class AbundanceForDiversity(Abundance):
    """Calculates metacommuntiy and subcommunity relative abundance
    components from a numpy.ndarray containing species counts
    """

    def __init__(
        self, counts: ndarray, subcommunity_names: Iterable[Union[str, int]]
    ) -> None:
        super().__init__(counts, subcommunity_names)
        self.metacommunity_abundance = self.make_metacommunity_abundance()
        self.unified_abundance_array = None

    def unify_abundance_array(self) -> None:
        """Creates one matrix containing all the abundance matrices:
        metacommunity, subcommunity, and normalized subcommunity.
        These matrices are still available as views on the unified
        data structure. (Because we are using basic slicing here, only
        one copy of the data will exist after garbage collection.)

        This allows for a major computational improvement in efficiency:
        The similarity matrix only has to be generated and used
        once (in the case where a pre-computed similarity matrix is not
        in RAM). That is, we make only one call to
        similarity.weighted_abundances(), in cases where generation of the
        similarity matrix is expensive.
        """
        self.unified_abundance_array = concatenate(
            (
                self.metacommunity_abundance,
                self.subcommunity_abundance,
                self.normalized_subcommunity_abundance,
            ),
            axis=1,
        )

    def get_unified_abundance_array(self):
        if self.unified_abundance_array is None:
            self.unify_abundance_array()
            self.metacommunity_abundance = self.unified_abundance_array[:, [0]]
            self.subcommunity_abundance = self.unified_abundance_array[
                :, 1 : (1 + self.num_subcommunities)
            ]
            self.normalized_subcommunity_abundance = self.unified_abundance_array[
                :, (1 + self.num_subcommunities) :
            ]
        return self.unified_abundance_array

    def premultiply_by(self, similarity):
        if similarity.is_expensive():
            all_ordinariness = similarity.self_similar_weighted_abundances(
                self.get_unified_abundance_array()
            )
            metacommunity_ordinariness = all_ordinariness[:, [0]]
            subcommunity_ordinariness = all_ordinariness[
                :, 1 : (1 + self.num_subcommunities)
            ]
            normalized_subcommunity_ordinariness = all_ordinariness[
                :, (1 + self.num_subcommunities) :
            ]
        else:
            metacommunity_ordinariness = similarity.self_similar_weighted_abundances(
                self.metacommunity_abundance
            )
            subcommunity_ordinariness = similarity.self_similar_weighted_abundances(
                self.subcommunity_abundance
            )
            normalized_subcommunity_ordinariness = (
                similarity.self_similar_weighted_abundances(
                    self.normalized_subcommunity_abundance
                )
            )
        return (
            metacommunity_ordinariness,
            subcommunity_ordinariness,
            normalized_subcommunity_ordinariness,
        )

    def make_metacommunity_abundance(self) -> ndarray:
        """Calculates the relative abundances in metacommunity.

        Returns
        -------
        A numpy.ndarray of shape (n_species, 1), where rows correspond
        to unique species and each row contains the relative abundance
        of the species in the metacommunity.
        """
        return self.subcommunity_abundance.sum(axis=1, keepdims=True)


def make_abundance(counts: Union[DataFrame, ndarray], for_diversity=True) -> Abundance:
    """Initializes a concrete subclass of Abundance.

    Parameters
    ----------
    counts:
        2-d array with one column per subcommunity, one row per species,
        where the elements are the species counts

    Returns
    -------
    An instance of a concrete subclass of Abundance.
    """
    if not for_diversity:
        specific_class = Abundance
    else:
        specific_class = AbundanceForDiversity
    if isinstance(counts, DataFrame):
        return specific_class(
            counts=counts.to_numpy(), subcommunity_names=counts.columns.to_list()
        )
    elif hasattr(counts, "shape"):
        if issparse(counts):
            raise TypeError("sparse abundance matrix not yet implemented")
        else:
            return specific_class(
                counts=counts, subcommunity_names=arange(counts.shape[1])
            )
    else:
        raise NotImplementedError(
            f"Type {type(counts)} is not supported for argument "
            "'counts'. Valid types include pandas.DataFrame or"
            "numpy.ndarray"
        )



from functools import cached_property
from numpy import ndarray

from greylock.abundance import Abundance
from greylock.similarity import Similarity


class Components:
    """Dispatches diversity components based on specified measure.
    If the similarity matrix is not the identity matrix, these
    will be similarity-sensitive diversity components."""

    def __init__(self, abundance: Abundance, similarity: Similarity) -> None:
        self.abundance = abundance

        """Create the ordinariness vectors by multipying the
        similarity matrix with each of the metacommunity abundance vector,
        the subcommunity abundance vectors, and the normalized
        subcommunity vectors.
        (See Leinster book* page 174 for discussion of ordinariness.
        * https://arxiv.org/pdf/2012.02113)
        Of course, for IdentitySimilarity, this multiplication would be
        a no-op (and thus is not actually performed).
        """
        (
            self.metacommunity_ordinariness,
            self.subcommunity_ordinariness,
            self.normalized_subcommunity_ordinariness,
        ) = self.abundance.premultiply_by(similarity)

        self.numerators = {
            **dict.fromkeys(["alpha", "gamma", "normalized_alpha"], 1),
            **dict.fromkeys(
                [
                    "beta",
                    "rho",
                    "normalized_beta",
                    "normalized_rho",
                    "beta_hat",
                    "rho_hat",
                ],
                self.metacommunity_ordinariness,
            ),
        }
        self.denominators = {
            **dict.fromkeys(
                ["alpha", "beta", "rho", "beta_hat", "rho_hat"],
                self.subcommunity_ordinariness,
            ),
            **dict.fromkeys(
                ["normalized_alpha", "normalized_beta", "normalized_rho"],
                self.normalized_subcommunity_ordinariness,
            ),
            "gamma": self.metacommunity_ordinariness,
        }



"""
Mock versions of ray functions for unit testing
"""

from collections import namedtuple
import math

counter = 0
results_store = {}


def get_next_token():
    global counter
    counter += 1
    return counter


def put(data):
    return data


RemoteWrap = namedtuple("RemoteWrap", ["remote"])


def remote(func):
    def remote(*args, **kwargs):
        token = get_next_token()
        result = func(*args, **kwargs)
        results_store[token] = result
        return token

    return RemoteWrap(remote=remote)


def wait(tokens):
    # Choose index of job to report as "ready".
    # Generally, don't choose the first one; clients
    # should not count on jobs finishing in order.
    ready_index = math.floor(len(tokens) / 3)
    ready_refs = [tokens[ready_index]]
    futures = tokens[:ready_index] + tokens[(ready_index + 1) :]
    return (ready_refs, futures)


def get_one(token):
    result = results_store[token]
    del results_store[token]
    return result


def get(tokens):
    if not hasattr(tokens, "__iter__"):
        return get_one(tokens)
    results = []
    for token in tokens:
        results.append(get_one(token))
    return results



"""Tests for diversity.metacommunity."""

from dataclasses import dataclass, field
from numpy import allclose, array, ndarray, identity, zeros, inf, maximum
from numpy.linalg import norm
from pandas import DataFrame, concat
from pandas.testing import assert_frame_equal
from pytest import mark, raises
from greylock.exceptions import InvalidArgumentError

from greylock.log import LOGGER
from greylock.abundance import Abundance
from greylock.similarity import (
    Similarity,
    SimilarityIdentity,
    SimilarityFromArray,
    SimilarityFromSymmetricFunction,
    SimilarityFromFunction,
)
from greylock import Metacommunity
from greylock.tests.base_tests.similarity_test import similarity_dataframe_3by3
from greylock.tests.base_tests.similarity_test import similarity_array_3by3_1

MEASURES = (
    "alpha",
    "rho",
    "beta",
    "gamma",
    "normalized_alpha",
    "normalized_rho",
    "normalized_beta",
    "rho_hat",
    "beta_hat",
)

subcommunity_names = ["subcommunity_1", "subcommunity_2"]
counts_3by2 = DataFrame([[1, 5], [3, 0], [0, 1]], columns=subcommunity_names)
counts_6by2 = DataFrame(
    [[1, 0], [1, 0], [1, 0], [0, 1], [0, 1], [0, 1]],
    columns=subcommunity_names,
)


@dataclass
class FrequencyMetacommunity6by2:
    description = "frequency-sensitive metacommunity; 6 species, 2 subcommunities"
    viewpoint: float = 0.0
    counts: DataFrame = field(default_factory=lambda: counts_6by2)
    similarity: None = None
    metacommunity_similarity: None = None
    subcommunity_similarity: None = None
    normalized_subcommunity_similarity: None = None
    subcommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": subcommunity_names,
                "viewpoint": [0.0, 0.0],
                "alpha": [6.0, 6.0],
                "rho": [1.0, 1.0],
                "beta": [1.0, 1.0],
                "gamma": [6.0, 6.0],
                "normalized_alpha": [3.0, 3.0],
                "normalized_rho": [0.5, 0.5],
                "normalized_beta": [2.0, 2.0],
                "rho_hat": [0.0, 0.0],
                "beta_hat": [1.0, 1.0],
            },
        )
    )
    metacommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": ["metacommunity"],
                "viewpoint": [0.0],
                "alpha": [6.0],
                "rho": [1.0],
                "beta": [1.0],
                "gamma": [6.0],
                "normalized_alpha": [3.0],
                "normalized_rho": [0.5],
                "normalized_beta": [2.0],
                "rho_hat": [0.0],
                "beta_hat": [1.0],
            },
            index=[0],
        )
    )


@dataclass
class FrequencyMetacommunity3by2:
    description = "frequency-sensitive metacommunity; 3 species, 2 subcommunities"
    viewpoint: float = 2.0
    counts: DataFrame = field(default_factory=lambda: counts_3by2)
    similarity: None = None
    metacommunity_similarity: None = None
    subcommunity_similarity: None = None
    normalized_subcommunity_similarity: None = None
    subcommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": subcommunity_names,
                "viewpoint": [2.0, 2.0],
                "alpha": [4.0, 2.30769231],
                "rho": [1.26315789, 1.16129032],
                "beta": [0.79166667, 0.86111111],
                "gamma": [2.66666667, 1.93548387],
                "normalized_alpha": [1.6, 1.38461538],
                "normalized_rho": [0.50526316, 0.69677419],
                "normalized_beta": [1.97916667, 1.43518519],
                "rho_hat": [0.263158, 0.161290],
                "beta_hat": [0.583333, 0.722222],
            }
        )
    )
    metacommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": ["metacommunity"],
                "viewpoint": [2.0],
                "alpha": [2.7777777777777777],
                "rho": [1.2],
                "beta": [0.8319209039548022],
                "gamma": [2.173913043478261],
                "normalized_alpha": [1.4634146341463414],
                "normalized_rho": [0.6050420168067228],
                "normalized_beta": [1.612461673236969],
                "rho_hat": [0.190840],
                "beta_hat": [0.659420],
            },
            index=[0],
        )
    )


@dataclass
class SimilarityMetacommunity6by2:
    description = "similarity-sensitive metacommunity; 6 species, 2 subcommunities"
    viewpoint: float = 0.0
    counts: ndarray = field(default_factory=lambda: counts_6by2)
    similarity: ndarray = field(
        default_factory=lambda: array(
            [
                [1.0, 0.5, 0.5, 0.7, 0.7, 0.7],
                [0.5, 1.0, 0.5, 0.7, 0.7, 0.7],
                [0.5, 0.5, 1.0, 0.7, 0.7, 0.7],
                [0.7, 0.7, 0.7, 1.0, 0.5, 0.5],
                [0.7, 0.7, 0.7, 0.5, 1.0, 0.5],
                [0.7, 0.7, 0.7, 0.5, 0.5, 1.0],
            ]
        )
    )
    metacommunity_similarity: ndarray = field(
        default_factory=lambda: array(
            [
                [0.68333333],
                [0.68333333],
                [0.68333333],
                [0.68333333],
                [0.68333333],
                [0.68333333],
            ]
        )
    )
    subcommunity_similarity: ndarray = field(
        default_factory=lambda: (
            array(
                [
                    [0.33333333, 0.35],
                    [0.33333333, 0.35],
                    [0.33333333, 0.35],
                    [0.35, 0.33333333],
                    [0.35, 0.33333333],
                    [0.35, 0.33333333],
                ],
            ),
        )
    )
    normalized_subcommunity_similarity: ndarray = field(
        default_factory=lambda: array(
            [
                [0.66666667, 0.7],
                [0.66666667, 0.7],
                [0.66666667, 0.7],
                [0.7, 0.66666667],
                [0.7, 0.66666667],
                [0.7, 0.66666667],
            ]
        )
    )
    subcommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": subcommunity_names,
                "viewpoint": [0.0, 0.0],
                "alpha": [3.0, 3.0],
                "rho": [2.05, 2.05],
                "beta": [0.487805, 0.487805],
                "gamma": [1.463415, 1.463415],
                "normalized_alpha": [1.5, 1.5],
                "normalized_rho": [1.025, 1.025],
                "normalized_beta": [0.97561, 0.97561],
                "rho_hat": [1.05, 1.05],
                "beta_hat": [-0.02439, -0.02439],
            }
        )
    )
    metacommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": ["metacommunity"],
                "viewpoint": [0.0],
                "alpha": [3.0],
                "rho": [2.05],
                "beta": [0.487805],
                "gamma": [1.463415],
                "normalized_alpha": [1.5],
                "normalized_rho": [1.025],
                "normalized_beta": [0.97561],
                "rho_hat": [1.05],
                "beta_hat": [-0.02439],
            },
            index=[0],
        )
    )


@dataclass
class SimilarityMetacommunity3by2:
    description = "similarity-sensitive metacommunity; 3 species, 2 subcommunities"
    viewpoint: float = 2.0
    counts: DataFrame = field(default_factory=lambda: counts_3by2)
    similarity: ndarray = field(default_factory=lambda: array(similarity_array_3by3_1))
    metacommunity_similarity: ndarray = field(
        default_factory=lambda: array([[0.76], [0.62], [0.22]])
    )
    subcommunity_similarity: ndarray = field(
        default_factory=lambda: (
            array(
                [
                    [0.25, 0.51],
                    [0.35, 0.27],
                    [0.07, 0.15],
                ]
            ),
        )
    )
    normalized_subcommunity_similarity: ndarray = field(
        default_factory=lambda: array(
            [
                [0.625, 0.85],
                [0.875, 0.45],
                [0.175, 0.25],
            ]
        )
    )
    subcommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": subcommunity_names,
                "viewpoint": [2.0, 2.0],
                "alpha": [3.07692308, 2.22222222],
                "rho": [1.97775446, 1.48622222],
                "beta": [0.50562394, 0.67284689],
                "gamma": [1.52671756, 1.49253731],
                "normalized_alpha": [1.23076923, 1.33333333],
                "normalized_rho": [0.79110178, 0.89173333],
                "normalized_beta": [1.26405985, 1.12141148],
                "rho_hat": [0.977754, 0.486222],
                "beta_hat": [0.011247877758913116, 0.345694],
            }
        )
    )
    metacommunity_results: DataFrame = field(
        default_factory=lambda: DataFrame(
            {
                "community": ["metacommunity"],
                "viewpoint": [2.0],
                "alpha": [2.5],
                "rho": [1.6502801833927663],
                "beta": [0.5942352817544037],
                "gamma": [1.5060240963855422],
                "normalized_alpha": [1.2903225806451613],
                "normalized_rho": [0.8485572790897555],
                "normalized_beta": [1.1744247216675028],
                "rho_hat": [0.608604],
                "beta_hat": [0.026811],
            },
            index=[0],
        )
    )


metacommunity_data = (
    FrequencyMetacommunity6by2(),
    FrequencyMetacommunity3by2(),
    SimilarityMetacommunity6by2(),
    SimilarityMetacommunity3by2(),
)


@mark.parametrize(
    "data, expected",
    zip(
        metacommunity_data,
        (
            SimilarityIdentity,
            SimilarityIdentity,
            SimilarityFromArray,
            SimilarityFromArray,
        ),
    ),
)
def test_metacommunity(data, expected):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    assert isinstance(metacommunity, Metacommunity)
    assert isinstance(metacommunity.abundance, Abundance)
    assert isinstance(metacommunity.similarity, expected)


@mark.parametrize("measure", MEASURES)
@mark.parametrize("data", metacommunity_data)
def test_metacommunity_diversity(data, measure):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    metacommunity_diversity = metacommunity.metacommunity_diversity(
        measure=measure, viewpoint=data.viewpoint
    )
    assert allclose(metacommunity_diversity, data.metacommunity_results[measure])


@mark.parametrize("measure", MEASURES)
@mark.parametrize("data", metacommunity_data)
def test_subcommunity_diversity(data, measure):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    subcommunity_diversity = metacommunity.subcommunity_diversity(
        measure=measure, viewpoint=data.viewpoint
    )
    assert allclose(subcommunity_diversity, data.subcommunity_results[measure])


def test_subcommunity_diversity_invalid_measure():
    with raises(InvalidArgumentError):
        Metacommunity(counts=counts_3by2).subcommunity_diversity(
            measure="omega", viewpoint=0
        )


@mark.parametrize("data", metacommunity_data)
def test_subcommunities_to_dataframe(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    subcommunities_df = metacommunity.subcommunities_to_dataframe(data.viewpoint)
    assert_frame_equal(subcommunities_df, data.subcommunity_results)


@mark.parametrize("data", metacommunity_data)
def test_metacommunities_to_dataframe(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    metacommunity_df = metacommunity.metacommunity_to_dataframe(
        viewpoint=data.viewpoint
    )
    assert_frame_equal(metacommunity_df, data.metacommunity_results)


@mark.parametrize("data", metacommunity_data)
def test_to_dataframe(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    expected = concat(
        [data.metacommunity_results, data.subcommunity_results]
    ).reset_index(drop=True)
    assert_frame_equal(metacommunity.to_dataframe(viewpoint=data.viewpoint), expected)


@mark.parametrize("data", metacommunity_data)
def test_select_measures(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    selected_measures = [
        "alpha",
        "gamma",
        "normalized_rho",
    ]
    expected_columns = selected_measures + ["community", "viewpoint"]
    df = metacommunity.to_dataframe(
        viewpoint=data.viewpoint, measures=selected_measures
    )
    for col in df:
        assert col in expected_columns
    for col in expected_columns:
        assert col in df


def test_effective_counts():
    """
    Test that:
    * passing None, an instance of SimilarityIdentity, or an identity matrix gives the same results
    * normalized_alpha for each subcommunity is appropriate for effective species count for each viewpoint:
      - raw species count at q = 0
      - 1/(proportion of largest species) at q = infinity
      - intermediate values of q yield intermediate values
    * giving the species some similiarity changes all the effective species counts
    """
    counts = DataFrame(
        {"A": [1, 2, 3, 4, 5], "B": [10, 5, 5, 0, 0]},
        index=["apple", "orange", "banana", "pear", "blueberry"],
    )
    viewpoints = [0, 1, 2, 3, 4, 99, inf]
    first_df = None
    for sim in [None, SimilarityIdentity(), identity(5)]:
        m = Metacommunity(counts, sim)
        df = m.to_dataframe(
            measures=["alpha", "normalized_alpha"], viewpoint=viewpoints
        )
        df.set_index(["community", "viewpoint"], inplace=True)
        if first_df is None:
            first_df = df
        else:
            for col in first_df:
                for ind in first_df.index:
                    assert df.loc[ind][col] == first_df.loc[ind][col]
        assert df.loc[("A", 0)]["normalized_alpha"] == 5.0
        assert df.loc[("A", inf)]["normalized_alpha"] == 3.0
        assert df.loc[("B", 0)]["normalized_alpha"] == 3.0
        assert df.loc[("B", inf)]["normalized_alpha"] == 2.0
        for community in ["A", "B"]:
            for i in range(1, len(viewpoints)):
                assert (
                    df.loc[(community, viewpoints[i - 1])]["normalized_alpha"]
                    >= df.loc[(community, viewpoints[i])]["normalized_alpha"]
                )
    sim = identity(5)
    for i, j, val in [
        (1, 0, 0.5),
        (1, 2, 0.4),
        (0, 2, 0.4),
        (3, 4, 0.1),
    ]:
        sim[i, j] = sim[j, i] = val
    m = Metacommunity(counts, sim)
    df = m.to_dataframe(viewpoint=viewpoints)
    df.set_index(["community", "viewpoint"], inplace=True)
    for col in first_df:
        for ind in first_df.index:
            assert df.loc[ind][col] < first_df.loc[ind][col]
    for community in ["A", "B"]:
        for i in range(1, len(viewpoints)):
            assert (
                df.loc[(community, viewpoints[i - 1])]["normalized_alpha"]
                >= df.loc[(community, viewpoints[i])]["normalized_alpha"]
            )


def test_property1():
    """
    Test elementary property 1 from L&C:
    Symmetry. Diversity is unchanged by the order in which the species happen to be listed.
    """
    X = DataFrame(
        {
            "red": [255, 245, 245, 213, 44],
            "green": [0, 108, 236, 227, 13],
            "blue": [0, 66, 66, 120, 92],
            "core": [100, 0, 0, 100, 0],
            "round": [90, 95, 20, 50, 90],
        },
        index=["apple", "orange", "banana", "pear", "blueberry"],
    )
    communities = DataFrame(
        {"bowl": [3, 1, 5, 1, 0], "fridge": [1, 4, 0, 1, 20]},
        index=["apple", "orange", "banana", "pear", "blueberry"],
    )

    def similarity_function(species_i, species_j):
        return 1 / (1 + norm(species_i - species_j) / 100)

    num_species = communities.shape[0]
    viewpoints = [0, 1, 2, 4, 88, inf]
    measures = [
        "alpha",
        "rho",
        "beta",
        "gamma",
        "normalized_alpha",
        "normalized_rho",
        "normalized_beta",
        "rho_hat",
    ]

    def get_result():
        metacommunity = Metacommunity(
            communities,
            similarity=SimilarityFromSymmetricFunction(
                similarity_function, X=X.to_numpy()
            ),
        )
        return metacommunity.to_dataframe(
            viewpoint=viewpoints, measures=measures
        ).set_index(["community", "viewpoint"])

    df1 = get_result()
    X = X.sort_index()
    communities = communities.sort_index()
    df2 = get_result()
    assert allclose(df1.to_numpy(), df2.to_numpy())


def test_property2():
    """
    Test elementary property 2 from L&C:
    Absent species. Diversity is unchanged by adding a new species of abundance 0
    """
    labels_2b = (
        "ladybug",
        "bee",
        "butterfly",
        "lobster",
        "fish",
        "turtle",
        "parrot",
        "llama",
        "orangutan",
    )
    no_species_2b = len(labels_2b)
    S_2b = identity(n=no_species_2b)
    # fmt: off
    S_2b[0][1:9] = (0.60, 0.55, 0.45, 0.25, 0.22, 0.23, 0.18, 0.16)  # ladybug
    S_2b[1][2:9] = (      0.60, 0.48, 0.22, 0.23, 0.21, 0.16, 0.14)  # bee
    S_2b[2][3:9] = (            0.42, 0.27, 0.20, 0.22, 0.17, 0.15)  # bu’fly
    S_2b[3][4:9] = (                  0.28, 0.26, 0.26, 0.20, 0.18)  # lobster
    S_2b[4][5:9] = (                        0.75, 0.70, 0.66, 0.63)  # fish
    S_2b[5][6:9] = (                              0.85, 0.70, 0.70)  # turtle
    S_2b[6][7:9] = (                                    0.75, 0.72)  # parrot
    S_2b[7][8:9] = (                                          0.85)  # llama
    pass                                                             # orangutan
    # fmt: on

    S_2b = maximum(S_2b, S_2b.transpose())
    counts = DataFrame({"Community 2b": [1, 1, 1, 1, 1, 1, 1, 1, 0]}, index=labels_2b)
    viewpoints = [0, 1, 2, 3, 4, 5, inf]
    metacommunity = Metacommunity(counts, similarity=S_2b)
    df1 = metacommunity.to_dataframe(viewpoint=viewpoints).set_index(
        ["community", "viewpoint"]
    )
    counts = counts[counts["Community 2b"] > 0]
    S_2b = S_2b[:-1, :-1]
    metacommunity = Metacommunity(counts, similarity=S_2b)
    df2 = metacommunity.to_dataframe(viewpoint=viewpoints).set_index(
        ["community", "viewpoint"]
    )
    assert allclose(df1.to_numpy(), df2.to_numpy())


def test_property3():
    """
    Test elementary property 3 from L&C:
    Identical species. If two species are identical, then merging them into one leaves the diversity unchanged.
    """
    labels = ["zucchini", "pumpkin", "eggplant", "aubergine"]
    no_species = len(labels)
    sim = identity(no_species)
    # fmt: off
    sim[0][1:4] = (0.5, 0.3, 0.3)
    sim[1][2:4] = (     0.3, 0.3)
    sim[2][3:4] = (          1.0)
    # fmt: on
    sim = maximum(sim, sim.T)
    counts = DataFrame(
        {"A": [3, 2, 4, 0], "B": [0, 4, 0, 5], "C": [1, 1, 1, 1]}, index=labels
    )
    viewpoints = [0, 1, 2, 3, 4, 5, 90, inf]
    measures = [
        "alpha",
        "rho",
        "beta",
        "gamma",
        "normalized_alpha",
        "normalized_rho",
        "normalized_beta",
        "rho_hat",
    ]
    metacommunity = Metacommunity(counts, sim)
    df1 = metacommunity.to_dataframe(viewpoint=viewpoints, measures=measures).set_index(
        ["community", "viewpoint"]
    )
    labels = labels[:-1]
    sim = sim[:-1, :-1]
    counts = DataFrame({"A": [3, 2, 4], "B": [0, 4, 5], "C": [1, 1, 2]}, index=labels)
    metacommunity = Metacommunity(counts, sim)
    df2 = metacommunity.to_dataframe(viewpoint=viewpoints, measures=measures).set_index(
        ["community", "viewpoint"]
    )
    assert allclose(df1.to_numpy(), df2.to_numpy(), equal_nan=True)


def test_figure_1():
    """
    Test that we get the results described in L&C pp. 482-483 (figure 1)
    """
    before_counts = array([[1], [3], [6]])
    before_sim = None
    naive_counts = array([[1], [3], [3], [3]])
    naive_sim = None
    nonnaive_counts = naive_counts
    nonnaive_sim = array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0.9], [0, 0, 0.9, 1]])
    before = Metacommunity(before_counts, before_sim)
    naive = Metacommunity(naive_counts, naive_sim)
    nonnaive = Metacommunity(nonnaive_counts, nonnaive_sim)
    for q in [0, 1, 2, 3, 4, 5, inf]:
        before_alpha = before.metacommunity_diversity(viewpoint=q, measure="alpha")
        naive_alpha = naive.metacommunity_diversity(viewpoint=q, measure="alpha")
        nonnaive_alpha = nonnaive.metacommunity_diversity(viewpoint=q, measure="alpha")
        assert (naive_alpha - before_alpha) >= 1.0
        assert (naive_alpha - nonnaive_alpha) > 0.9
        assert naive_alpha <= 4.0
        assert before_alpha <= 3.0
        assert (nonnaive_alpha - before_alpha) < 0.2



"""Tests for diversity.abundance."""

from dataclasses import dataclass, field
from numpy import allclose, array, ndarray
from pandas import DataFrame
from pytest import fixture, mark, raises
from scipy.sparse import coo_array  # type: ignore[import]

from greylock.abundance import (
    AbundanceForDiversity,
    make_abundance,
)


def test_sparse():
    sparse_abundance = coo_array(
        (array([4, 5, 7, 9]), (array([0, 3, 1, 0]), array([0, 3, 1, 2]))),
        shape=(4, 4),
    )
    with raises(TypeError):
        make_abundance(sparse_abundance)


def counts_array_3by2():
    return array([[2, 4], [3, 0], [0, 1]])


counts_dataframe_3by2 = DataFrame(counts_array_3by2())


@dataclass
class AbundanceExclusiveSpecies:
    description: str = "2 subcommunities; both contain exclusive species"
    counts: ndarray = field(default_factory=counts_array_3by2)
    subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [[2 / 10, 4 / 10], [3 / 10, 0 / 10], [0 / 10, 1 / 10]]
        )
    )
    metacommunity_abundance: ndarray = field(
        default_factory=lambda: array([[6 / 10], [3 / 10], [1 / 10]])
    )
    subcommunity_normalizing_constants: ndarray = field(
        default_factory=lambda: array([5 / 10, 5 / 10])
    )
    normalized_subcommunity_abundance: ndarray = field(
        default_factory=lambda: array([[2 / 5, 4 / 5], [3 / 5, 0 / 5], [0 / 5, 1 / 5]])
    )


@dataclass
class AbundanceSingleExclusiveSpecies:
    description: str = "2 subcommunities; one contains exclusive species"
    counts: ndarray = field(default_factory=lambda: array([[2, 4], [3, 0], [5, 1]]))
    subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [[2 / 15, 4 / 15], [3 / 15, 0 / 15], [5 / 15, 1 / 15]]
        )
    )
    metacommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [6 / 15],
                [3 / 15],
                [6 / 15],
            ]
        )
    )
    subcommunity_normalizing_constants: ndarray = field(
        default_factory=lambda: array([10 / 15, 5 / 15])
    )
    normalized_subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [[2 / 10, 4 / 5], [3 / 10, 0 / 5], [5 / 10, 1 / 5]]
        )
    )


@dataclass
class AbundanceNoExclusiveSpecies:
    description: str = "2 communities; neither contain exclusive species"
    counts: ndarray = field(
        default_factory=lambda: array(
            [
                [2, 4],
                [3, 1],
                [1, 5],
            ],
        )
    )
    subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 16, 4 / 16],
                [3 / 16, 1 / 16],
                [1 / 16, 5 / 16],
            ]
        )
    )
    metacommunity_abundance: ndarray = field(
        default_factory=lambda: array([[6 / 16], [4 / 16], [6 / 16]])
    )
    subcommunity_normalizing_constants: ndarray = field(
        default_factory=lambda: array([6 / 16, 10 / 16])
    )
    normalized_subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 6, 4 / 10],
                [3 / 6, 1 / 10],
                [1 / 6, 5 / 10],
            ]
        )
    )


@dataclass
class AbundanceMutuallyExclusive:
    description: str = "2 mutually exclusive communities"
    counts: ndarray = field(
        default_factory=lambda: array(
            [
                [2, 0],
                [3, 0],
                [0, 1],
                [0, 4],
            ],
        )
    )
    subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 10, 0 / 10],
                [3 / 10, 0 / 10],
                [0 / 10, 1 / 10],
                [0 / 10, 4 / 10],
            ]
        )
    )
    metacommunity_abundance: ndarray = field(
        default_factory=lambda: array([[2 / 10], [3 / 10], [1 / 10], [4 / 10]])
    )
    subcommunity_normalizing_constants: ndarray = field(
        default_factory=lambda: array([5 / 10, 5 / 10])
    )
    normalized_subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 5, 0 / 5],
                [3 / 5, 0 / 5],
                [0 / 5, 1 / 5],
                [0 / 5, 4 / 5],
            ]
        )
    )


@dataclass
class AbundanceOneSubcommunity:
    description: str = "one community"
    counts: ndarray = field(
        default_factory=lambda: array(
            [
                [2],
                [5],
                [3],
            ],
        )
    )
    subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 10],
                [5 / 10],
                [3 / 10],
            ]
        )
    )
    metacommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 10],
                [5 / 10],
                [3 / 10],
            ]
        )
    )
    subcommunity_normalizing_constants: ndarray = field(
        default_factory=lambda: array([10 / 10])
    )
    normalized_subcommunity_abundance: ndarray = field(
        default_factory=lambda: array(
            [
                [2 / 10],
                [5 / 10],
                [3 / 10],
            ]
        )
    )


@mark.parametrize(
    "counts, expected",
    [
        (counts_array_3by2(), AbundanceForDiversity),
        (counts_dataframe_3by2, AbundanceForDiversity),
    ],
)
def test_make_abundance(counts, expected):
    abundance = make_abundance(counts=counts)
    assert isinstance(abundance, expected)


def test_make_abundance_not_implemented():
    with raises(NotImplementedError):
        make_abundance(counts=1)


@mark.parametrize(
    "test_case",
    [
        AbundanceExclusiveSpecies(),
        AbundanceSingleExclusiveSpecies(),
        AbundanceNoExclusiveSpecies(),
        AbundanceMutuallyExclusive(),
        AbundanceOneSubcommunity(),
    ],
)
class TestAbundance:
    def test_make_subcommunity_abundance(self, test_case):
        abundance = make_abundance(counts=test_case.counts)
        assert allclose(
            abundance.subcommunity_abundance,
            test_case.subcommunity_abundance,
        )

    def test_metacommunity_abundance(self, test_case):
        abundance = make_abundance(counts=test_case.counts)
        assert allclose(
            abundance.metacommunity_abundance, test_case.metacommunity_abundance
        )

    def test_subcommunity_normalizing_constants(self, test_case):
        abundance = make_abundance(counts=test_case.counts)
        assert allclose(
            abundance.subcommunity_normalizing_constants,
            test_case.subcommunity_normalizing_constants,
        )

    def test_normalized_subcommunity_abundance(self, test_case):
        abundance = make_abundance(counts=test_case.counts)
        assert allclose(
            abundance.normalized_subcommunity_abundance,
            test_case.normalized_subcommunity_abundance,
        )



"""Module for calculating weighted subcommunity and metacommunity
similarities.

At the heart of calculating diversity with similarity is the matrix multiplication
    Z @ p
where Z is the similarity matrix and p is a vector (for a single unified
community) or matrix (for subcommunities of a metacommunity) representing
abundances.
( See the Leinster and Cobbold paper, equation (1).)
The Similarity class is responsible for premultiplying abundances by a
similarity matrix. If compute resources were infinite, this would just
be a matrix multiplication. However, the similarity matrix may be extremely
large and/or expensive to calculate. Various subclasses of Similarity use
various strategies for dealing with the generation and use of this
similarity matrix in various ways. For example:
1) Ignore similarity and do similarity-naive diversity calculations:
  Equivalent to Z being I, and the multiplication is a no-op.
  Implemented by SimilarityIdentity
2) The computionally easy/naive case: Z can be practically be calculated
  and kept in memory as a NumPy array: SimilarityFromArray
3) The similarity matrix is calculated in a separate job and stored in a
   file: SimilarityFromFile
4) The similarity matrix is too large to pre-calculate, and needs to be
   generated on the fly as the matrix multiplication is performed:
   SimilarityFromFunction

In some of these classes (SimilarityFromArray, SimilarityFromArray) the
similarity matrix exists explicitly and in others (SimilarityIdentity,
SimilarityFromFunction) it is implicit, or only exists piecemeal at a time.
In any case, one can caputure the similarity matrix that would be used
by passing similarities_out (a NumPy array of appropriate shape)
to the weighted_abundances method. Of course, in some cases this may be
silly (SimilarityIdentity) or unwise (SimilarityFromFunction with a large
species count), but this functionality is implemented in all concrete
Similarity classes to maintain consistency and correctness.

Classes
-------
Similarity
    Abstract base class for relative abundance-weighted species
    similarities.
SimilarityIdentity
    No-op version (equivalent to Z = I) for calculating similarity-
    insensitive diversity.
SimilarityFromDataFrame
    Implements Similarity by storing similarities in a pandas DataFrame.
SimilarityFromArray
    Implements Similarity by storing similarities in a numpy ndarray or
    memmap.
SimilarityFromFile
    Implements Similarity by reading similarities from a csv or tsv
    file.
SimilarityFromFunction:
    Implements Similarity by calculating pairwise similarities with a
    callable function. See `ray.py` for a parallelized version of this.
SimilarityFromSymmetricFunction
    Optimized version of SimilarityFunction that skips unneeded calculations.
    Only valid when Z[i,j] == Z[j,i] and Z[i,i] == 1.0 for all i, j.
    See `ray.py` for a parallelized version of this.
IntersetSimilarityFromFunction
    Not relevant to diversity calculations.
    Used to calcuate a similarity matrix in which the rows and columns
    correspond to two separate sets of species, to probe a community for
    similarity to some other set of species of interest.
    Calculates pairwise similarities with a callable function.
    See `ray.py` for a parallelized version of this.


Functions
---------
weighted_similarity_chunk_nonsymmetric
     Utility funciton for calculation of chunks of a similarity matrix
weighted_similarity_chunk_symmetric
     Utility funciton for calculation of chunks of a symmetric similarity matrix
"""

from abc import ABC, abstractmethod
from typing import Callable, Union, Tuple
from pathlib import Path
from numpy import ndarray, empty, concatenate, float64, vstack
from numpy import zeros, identity
from pandas import DataFrame, read_csv
from scipy.sparse import spmatrix, issparse  # type: ignore[import]
from greylock.exceptions import InvalidArgumentError


class Similarity(ABC):
    """Root superclass for classes computing weighted similarities."""

    def __init__(self, similarities_out: Union[ndarray, None] = None):
        """
        Parameters
        ----------
        similarities_out:
           If not None, a numpy array into which the similarity matrix
           will be written.

        """
        self.similarities_out = similarities_out

    @abstractmethod
    def weighted_abundances(
        self,
        relative_abundances: Union[ndarray, spmatrix],
    ) -> ndarray:
        """Calculates weighted sums of similarities to each species.
        Parameters
        ________
        relative_abundances:
           2-d array of shape (n_species, n_communities), where rows
           rows correspond to unique species, columns correspond to
           communities, and each element is some type of count
           (absolute or normalized).

        Returns
        -------
        A 2-d numpy.ndarray of shape (n_species, n_communities), where
        rows correspond to unique species, columns correspond to
        (meta-/sub-) communities and each element is a sum of
        similarities of all species to the row's species weighted by
        their relative abundances in the respective communities.
        """
        pass

    def self_similar_weighted_abundances(
        self, relative_abundances: Union[ndarray, spmatrix]
    ) -> ndarray:
        return self.weighted_abundances(relative_abundances)

    def is_expensive(self):
        return False

    def __matmul__(self, abundance):
        return abundance.premultiply_by(self)


class SimilarityIdentity(Similarity):
    def weighted_abundances(self, relative_abundance):
        if self.similarities_out is not None:
            # create the identity matrix in place
            self.similarities_out.fill(0.0)
            for i in range(relative_abundance.shape[0]):
                self.similarities_out[i, i] = 1.0
        return relative_abundance


class SimilarityFromArray(Similarity):
    """Implements Similarity using similarities stored in a numpy
    ndarray."""

    def __init__(
        self,
        similarity: Union[ndarray, spmatrix],
        similarities_out: Union[ndarray, None] = None,
    ):
        super().__init__(similarities_out)
        self.similarity = similarity

    def weighted_abundances(
        self,
        relative_abundance: Union[ndarray, spmatrix],
    ) -> ndarray:
        if self.similarities_out is not None:
            self.similarities_out[:, :] = self.similarity
        return self.similarity @ relative_abundance

    def self_similar_weighted_abundances(
        self, relative_abundances: Union[ndarray, spmatrix]
    ) -> ndarray:
        if self.similarity.shape[0] == self.similarity.shape[1]:
            return self.weighted_abundances(relative_abundances)
        else:
            raise InvalidArgumentError("Similarity matrix must be square")


class SimilarityFromDataFrame(SimilarityFromArray):
    """Implements Similarity using similarities stored in pandas
    dataframe."""

    def __init__(
        self, similarity: DataFrame, similarities_out: Union[ndarray, None] = None
    ):
        super().__init__(similarity.to_numpy(), similarities_out)


class SimilarityFromFile(Similarity):
    """Implements Similarity by using similarities stored in file.

    Similarity matrix rows are read from the file one chunk at a time.
    The size of chunks can be specified in numbers of rows to control
    memory load.
    """

    def __init__(
        self,
        similarity_file_path: Union[str, Path],
        chunk_size: int = 100,
        similarities_out: Union[ndarray, None] = None,
    ) -> None:
        """
        Parameters
        ----------
        similarity_file_path:
            Path to a file containing a pairwise similarity matrix of
            shape (n_species, n_species). The file should have a header
            that denotes the unique species names.
        chunk_size:
            Number of rows to read from similarity matrix at a time.
        """
        super().__init__(similarities_out)
        self.path = Path(similarity_file_path)
        self.chunk_size = chunk_size

    def weighted_abundances(
        self,
        relative_abundance: Union[ndarray, spmatrix],
    ) -> ndarray:
        weighted_abundances = empty(relative_abundance.shape, dtype=float64)
        with read_csv(
            self.path,
            chunksize=self.chunk_size,
            sep=None,
            engine="python",
            dtype=float64,
        ) as similarity_matrix_chunks:
            i = 0
            for chunk in similarity_matrix_chunks:
                chunk_as_numpy = chunk.to_numpy()
                weighted_abundances[i : i + self.chunk_size, :] = (
                    chunk_as_numpy @ relative_abundance
                )
                if self.similarities_out is not None:
                    self.similarities_out[i : i + self.chunk_size, :] = chunk_as_numpy
                i += self.chunk_size
        return weighted_abundances

    def is_expensive(self):
        return True


class IntersetSimilarityFromFile(SimilarityFromFile):
    def weighted_abundances(
        self,
        relative_abundance: Union[ndarray, spmatrix],
    ) -> ndarray:
        with read_csv(
            self.path,
            chunksize=self.chunk_size,
            sep=None,
            engine="python",
            dtype=float64,
        ) as similarity_matrix_chunks:
            weighted_abundance_chunks = []
            for j, chunk in enumerate(similarity_matrix_chunks):
                chunk_as_numpy = chunk.to_numpy()
                if self.similarities_out is not None:
                    self.similarities_out[
                        j * self.chunk_size : (j + 1) * self.chunk_size, :
                    ] = chunk_as_numpy
                weighted_abundance_chunks.append(chunk_as_numpy @ relative_abundance)
        return concatenate(weighted_abundance_chunks)

    def self_similar_weighted_abundances(
        self, relative_abundance: Union[ndarray, spmatrix]
    ):
        raise InvalidArgumentError(
            "Inappropriate similarity class for diversity measures"
        )


def weighted_similarity_chunk_nonsymmetric(
    similarity: Callable,
    X: Union[ndarray, DataFrame],
    Y: Union[ndarray, DataFrame, None],
    relative_abundance: ndarray,
    chunk_size: int,
    chunk_index: int,
    return_Z: bool = True,
) -> Tuple[int, ndarray, Union[ndarray, None]]:
    """
    Calculates some rows of the matrix product of Z @ p,
    where Z is not given explicitly but rather each entry
    Z[i,j] is calculated by a function.

    It need not be the case that Z[i,j] == Z[j,i], or even
    that Z is square; in such symmetric cases consider using
    weighted_similarity_chunk_symmetric.

    Parameters
    ----------
    similarity : a callable which, given row i of X and row j of Y,
      calculates Z[i,j].
    X: a NumPy array or pandas DataFrame whose rows contain feature
      values for a given species.
    Y: If this is None, then Z[i,j] is the similarity between species
      i and j of the same list of species.
      If Y is given, then Y's rows contain feature values for a different
      set of species than X's rows.
    relative_abundance: a NumPy array with rows corresponding to species,
      columns correspond to (sub-)communities; the p in Z @ p.
    chunk_size: How many rows to calculate.
    chunk_index: The starting row index of the chunk to be calculated.
    return_Z: Whether to return the calculated rows of Z.

    Returns
    ------
    A tuple of 3 items.
    (0) chunk_index - int
      When this is a remote task, the chunks may be returned out of
      order. We return the chunk_index to indicate what chunk this was for,
      so that the caller can sort the resulting chunks correctly.
    (1) result - NumPy array
        This is chunk_size rows of the resulting weighted abundance array.
    (2) similarities_chunk - NumPy array (optional)
        If the caller wants to capture the Z that was used, this will be chunk_size
        rows of Z.
    """

    def enum_helper(X):
        if type(X) == DataFrame:
            return X.itertuples()
        return X

    if Y is None:
        Y = X
    chunk = X[chunk_index : chunk_index + chunk_size]
    similarities_chunk = empty(shape=(chunk.shape[0], Y.shape[0]))
    for i, row_i in enumerate(enum_helper(chunk)):
        for j, row_j in enumerate(enum_helper(Y)):
            similarities_chunk[i, j] = similarity(row_i, row_j)
    result = similarities_chunk @ relative_abundance
    if return_Z:
        return chunk_index, result, similarities_chunk
    else:
        return chunk_index, result, None


class SimilarityFromSymmetricFunction(Similarity):
    """
    Calculate a similarity matrix on the fly, given feature vectors for each
    species and a function to calculate similarity from feature vectors.

    This assumes that
    a) the similarity between any species and itself is 1.0.
    b) similarity is symmetric: that is, Z[i,j] == Z[j,i]] always by definition.

    Note that condition (b) need not be true and all the Leinster & Cobbald math
    is still valid (see L&C paper). If these conditions do not hold, use the
    next class: SimilarityFromFunction. But if you can use this class, you get the
    obvious speed-up.

    N.B. that these calculations can be parallelized; see ray.py for
    parallelization using the Ray package.
    """

    def __init__(
        self,
        func: Callable,
        X: Union[ndarray, DataFrame],
        chunk_size: int = 100,
        similarities_out: Union[ndarray, None] = None,
    ):
        """
        Parameters
        ----------
        func:
            A Callable that calculates similarity between a pair of species.
            Must take two rows from X as input as its arguments, and return
            a numeric similarity value between 0.0 and 1.0.
        X:
          Each row contains the features values for a given species.
        chunk_size:
            Number of rows in similarity matrix to calculate at a time.
        """
        super().__init__(similarities_out)
        self.func = func
        self.X = X
        self.chunk_size = chunk_size

    def is_expensive(self):
        return True

    def weighted_abundances(
        self,
        abundance: Union[ndarray, spmatrix],
    ) -> ndarray:
        result = abundance.copy()
        if self.similarities_out is not None:
            self.similarities_out.fill(0.0)
        for chunk_index in range(0, self.X.shape[0], self.chunk_size):
            _, chunk, similarities = weighted_similarity_chunk_symmetric(
                similarity=self.func,
                X=self.X,
                relative_abundance=abundance,
                chunk_size=self.chunk_size,
                chunk_index=chunk_index,
                return_Z=(self.similarities_out is not None),
            )
            result = result + chunk
            if self.similarities_out is not None:
                self.similarities_out[
                    chunk_index : chunk_index + self.chunk_size, :
                ] = similarities
        if self.similarities_out is not None:
            self.similarities_out += self.similarities_out.T
            identity_chunk = identity(self.chunk_size)
            for chunk_index in range(0, self.X.shape[0], self.chunk_size):
                if chunk_index + self.chunk_size <= self.similarities_out.shape[0]:
                    self.similarities_out[
                        chunk_index : chunk_index + self.chunk_size,
                        chunk_index : chunk_index + self.chunk_size,
                    ] += identity_chunk
                else:
                    for i in range(chunk_index, self.similarities_out.shape[0]):
                        self.similarities_out[i, i] = 1.0
        return result


def weighted_similarity_chunk_symmetric(
    similarity: Callable,
    X: Union[ndarray, DataFrame],
    relative_abundance: ndarray,
    chunk_size: int,
    chunk_index: int,
    return_Z: bool = True,
) -> Tuple[int, ndarray, Union[ndarray, None]]:
    """
    Calculates partial results of the matrix product of Z @ p,
    where Z is not given explicitly but rather each entry
    Z[i,j] is calculated by a function. Assumes that
    Z[i,j] == Z[j,i] always.

    Parameters
    ----------
    similarity : a callable which, given row i of X and row j of Y,
      calculates Z[i,j].
    X: a NumPy array or pandas DataFrame whose rows contain feature
      values for a given species.
    relative_abundance: a NumPy array with rows corresponding to species,
      columns correspond to (sub-)communities; the p in Z @ p.
    chunk_size: How many rows of Z to calculate.
    chunk_index: The starting row index of the chunk of Z to be calculated
      and used.
    return_Z: Whether to return the calculated rows of Z.

    Returns
    ------
    A tuple of 3 items.
    (0) similarities_chunk - NumPy array (optional)
        If the caller wants to capture the Z that was used, this will be chunk_size
        rows of a matrix such that the upper triangle contains values of Z.
        (Lower triangle and diagonal values will be all zero.)
    (1) result - NumPy array
        This is a partial result of the resulting weighted abundance array.
    (0) chunk_index - int
      When this is a remote task, the chunks may be returned out of
      order. We return the chunk_index to indicate what chunk this was for,
      so that the caller can sort the resulting chunks correctly.
    """

    def enum_helper(X, start_index=0):
        if type(X) == DataFrame:
            return X.iloc[start_index:].itertuples()
        return X[start_index:]

    chunk = X[chunk_index : chunk_index + chunk_size]
    similarities_chunk = zeros(shape=(chunk.shape[0], X.shape[0]))
    for i, row_i in enumerate(enum_helper(chunk)):
        for j, row_j in enumerate(enum_helper(X, chunk_index + i + 1)):
            similarities_chunk[i, i + j + chunk_index + 1] = similarity(row_i, row_j)
    rows_result = similarities_chunk @ relative_abundance
    rows_after_count = max(0, relative_abundance.shape[0] - (chunk_index + chunk_size))
    rows_result = vstack(
        (
            zeros(shape=(chunk_index, relative_abundance.shape[1])),
            rows_result,
            zeros(
                shape=(
                    rows_after_count,
                    relative_abundance.shape[1],
                )
            ),
        )
    )
    relative_abundance = relative_abundance[chunk_index : chunk_index + chunk_size]
    cols_result = similarities_chunk.T @ relative_abundance
    result = rows_result + cols_result
    if return_Z:
        return chunk_index, result, similarities_chunk
    else:
        return chunk_index, result, None


class SimilarityFromFunction(SimilarityFromSymmetricFunction):
    def weighted_abundances(
        self,
        relative_abundance: Union[ndarray, spmatrix],
    ):
        weighted_similarity_chunks = []
        for chunk_index in range(0, self.X.shape[0], self.chunk_size):
            _, result, similarities = weighted_similarity_chunk_nonsymmetric(
                self.func,
                self.X,
                self.get_Y(),
                relative_abundance,
                self.chunk_size,
                chunk_index,
                (self.similarities_out is not None),
            )
            weighted_similarity_chunks.append(result)
            if self.similarities_out is not None:
                self.similarities_out[
                    chunk_index : chunk_index + result.shape[0], :
                ] = similarities
        return concatenate(weighted_similarity_chunks)

    def get_Y(self):
        return None


class IntersetSimilarityFromFunction(SimilarityFromFunction):
    def __init__(
        self,
        func: Callable,
        X: Union[ndarray, DataFrame],
        Y: Union[ndarray, DataFrame],
        chunk_size: int = 100,
        similarities_out: Union[ndarray, None] = None,
    ):
        """
        Parameters
        ----------
        func:
            A Callable that calculates similarity between a pair of species.
            Must take a row from X and a row from Y as input as its arguments, and return
            a numeric similarity value between 0.0 and 1.0.
        X:
          Each row contains the features values for a given species in set A.
        Y:
          Each row contains the features values for a given species in set B.
        chunk_size:
            Number of rows in similarity matrix to calculate at a time.
        """
        super().__init__(func, X, chunk_size, similarities_out)
        self.Y = Y

    def get_Y(self):
        return self.Y

    def self_similar_weighted_abundances(
        self, relative_abundance: Union[ndarray, spmatrix]
    ):
        raise InvalidArgumentError(
            "Inappropriate similarity class for diversity measures"
        )



"""Package for calculating Similarity-sensitive diversities.

The diversity measures that can be calculated are described rigorously
in https://arxiv.org/abs/1404.6520.

See greylock.metacommunity.Metacommunity
"""

from greylock.metacommunity import Metacommunity



"""Module for metacommunity and subcommunity diversity measures.

Classes
-------
Metacommunity
    Represents a metacommunity made up of subcommunities and computes
    metacommunity and subcommunity diversity measures.
"""

from typing import Callable
from collections.abc import Iterable
from typing import Union

from pandas import DataFrame, Index, Series, concat
from numpy import array, atleast_1d, broadcast_to, divide, zeros, ndarray
from greylock.exceptions import InvalidArgumentError

from greylock.abundance import make_abundance
from greylock.similarity import Similarity, SimilarityFromArray, SimilarityIdentity
from greylock.components import Components
from greylock.powermean import power_mean


class Metacommunity:
    similarity: Similarity
    """Creates diversity components and calculates diversity measures.

    A community consists of a set of species, each of which may appear
    any (non-negative) number of times. A metacommunity consists of one
    or more subcommunities and can be represented by the number of
    appearances of each species in each of the subcommunities that the
    species appears in.
    """

    MEASURES = (
        "alpha",
        "rho",
        "beta",
        "gamma",
        "normalized_alpha",
        "normalized_rho",
        "normalized_beta",
        "rho_hat",
        "beta_hat",
    )

    def __init__(
        self,
        counts: Union[DataFrame, ndarray],
        similarity: Union[ndarray, Similarity, None] = None,
    ) -> None:
        """
        Parameters
        ----------
        counts:
            2-d array with one column per subcommunity, one row per
            species, containing the count of each species in the
            corresponding subcommunities.
        similarity:
            For small datasets this can be the similarity matrix as
            an n-by-n numpy.ndarray.
            For larger datasets, various subclasses of Similarity
            provide the similarity matrix in various memory- and compute-
            efficient way.
            If None is given here, the diversity measures calculated will
            be frequency-sensitive only, not similarity-sensitive.
        """
        self.counts = counts
        self.abundance = make_abundance(counts=counts)
        if similarity is None:
            self.similarity = SimilarityIdentity()
        elif isinstance(similarity, ndarray):
            self.similarity = SimilarityFromArray(similarity=similarity)
        else:
            self.similarity = similarity
        self.components = Components(
            abundance=self.abundance, similarity=self.similarity
        )

    def subcommunity_diversity(self, viewpoint: float, measure: str) -> ndarray:
        """Calculates subcommunity diversity measures.

        Parameters
        ----------
        viewpoint:
            Viewpoint parameter for diversity measure.
        measure:
            Name of the diversity measure. Valid measures include:
            "alpha", "rho", "beta", "gamma", "normalized_alpha",
            "normalized_rho", and "normalized_beta"

        Returns
        -------
        A numpy.ndarray with a diversity measure for each subcommunity.
        """
        if measure not in self.MEASURES:
            raise (
                InvalidArgumentError(
                    f"Invalid measure '{measure}'. "
                    "Argument 'measure' must be one of: "
                    f"{', '.join(self.MEASURES)}"
                )
            )
        numerator = self.components.numerators[measure]
        denominator = self.components.denominators[measure]
        if measure == "gamma":
            denominator = broadcast_to(
                denominator,
                self.abundance.normalized_subcommunity_abundance.shape,
            )
        community_ratio = divide(
            numerator,
            denominator,
            out=zeros(denominator.shape),
            where=denominator != 0,
        )
        diversity_measure = power_mean(
            order=1 - viewpoint,
            weights=self.abundance.normalized_subcommunity_abundance,
            items=community_ratio,
            atol=self.abundance.min_count,
        )
        if measure in {"beta", "normalized_beta"}:
            return 1 / diversity_measure

        if measure in {"rho_hat"} and self.counts.shape[1] > 1:
            N = self.counts.shape[1]
            return (diversity_measure - 1) / (N - 1)

        if measure in {"beta_hat"} and self.counts.shape[1] > 1:
            N = self.counts.shape[1]
            return ((N / diversity_measure) - 1) / (N - 1)

        return diversity_measure

    def metacommunity_diversity(self, viewpoint: float, measure: str) -> ndarray:
        """Calculates metacommunity diversity measures.

        Parameters
        ----------
        viewpoint:
            Viewpoint parameter for diversity measure.
        measure:
            Name of the diversity measure. Valid measures include:
            "alpha", "rho", "beta", "gamma", "normalized_alpha",
            "normalized_rho", and "normalized_beta"

        Returns
        -------
        A numpy.ndarray containing the metacommunity diversity measure.
        """
        subcommunity_diversity = self.subcommunity_diversity(viewpoint, measure)
        return power_mean(
            1 - viewpoint,
            self.abundance.subcommunity_normalizing_constants,
            subcommunity_diversity,
        )

    def subcommunities_to_dataframe(self, viewpoint: float, measures=MEASURES):
        """Table containing all subcommunity diversity values.

        Parameters
        ----------
        viewpoint:
            Affects the contribution of rare species to the diversity
            measure. When viewpoint = 0 all species (rare or frequent)
            contribute equally. When viewpoint = infinity, only the
            single most frequent species contribute.

        Returns
        -------
        A pandas.DataFrame containing all subcommunity diversity
        measures for a given viewpoint
        """
        df = DataFrame(
            {
                measure: self.subcommunity_diversity(viewpoint, measure)
                for measure in measures
            }
        )
        df.insert(0, "viewpoint", viewpoint)
        df.insert(0, "community", Series(self.abundance.subcommunities_names))
        return df

    def metacommunity_to_dataframe(self, viewpoint: float, measures=MEASURES):
        """Table containing all metacommunity diversity values.

        Parameters
        ----------
        viewpoint:
            Affects the contribution of rare species to the diversity
            measure. When viewpoint = 0 all species (rare or frequent)
            contribute equally. When viewpoint = infinity, only the
            single most frequent species contributes.

        Returns
        -------
        A pandas.DataFrame containing all metacommunity diversity
        measures for a given viewpoint
        """
        df = DataFrame(
            {
                measure: self.metacommunity_diversity(viewpoint, measure)
                for measure in measures
            },
            index=Index(["metacommunity"], name="community"),
        )
        df.insert(0, "viewpoint", viewpoint)
        df.reset_index(inplace=True)
        return df

    def to_dataframe(self, viewpoint: Union[float, Iterable[float]], measures=MEASURES):
        """Table containing all metacommunity and subcommunity diversity
        values.

        Parameters
        ----------
        viewpoint:
            Affects the contribution of rare species to the diversity
            measure. When viewpoint = 0 all species (rare or frequent)
            contribute equally. When viewpoint = infinity, only the
            single most frequent species contributes.

        Returns
        -------
        A pandas.DataFrame containing all metacommunity and subcommunity
        diversity measures for a given viewpoint
        """
        dataframes = []
        for q in atleast_1d(array(viewpoint)):
            dataframes.append(
                self.metacommunity_to_dataframe(viewpoint=q, measures=measures)
            )
            dataframes.append(
                self.subcommunities_to_dataframe(viewpoint=q, measures=measures)
            )
        return concat(dataframes).reset_index(drop=True)



"""Main module for executing diversity on command-line.

Functions
---------
main
    Calculates diversities according to command-line specifications.
"""

from sys import argv
from platform import python_version
from logging import captureWarnings, getLogger
from numpy import int64

from pandas import read_csv

from greylock.log import LOG_HANDLER, LOGGER
from greylock import Metacommunity
from greylock.similarity import SimilarityFromFile
from greylock.parameters import configure_arguments

# Ensure warnings are handled properly.
captureWarnings(True)
getLogger("py.warnings").addHandler(LOG_HANDLER)


def main(args):
    """Calculates diversity from species counts and similarities.

    Parameters
    ----------
    args: argparse.Namespace
        Return object of argparse.ArgumentParser object created by
        diversity.parameters.configure_arguments and applied to command
        line arguments.
    """
    LOGGER.setLevel(args.log_level)
    LOGGER.info(" ".join([f"python{python_version()}", *argv]))
    LOGGER.debug(f"args: {args}")

    counts = read_csv(args.input_filepath, sep=None, engine="python", dtype=int64)
    LOGGER.debug(f"data: {counts}")
    similarity = SimilarityFromFile(args.similarity, args.chunk_size)
    metacommunity = Metacommunity(
        counts=counts,
        similarity=similarity,
    )
    community_views = metacommunity.to_dataframe(viewpoint=args.viewpoint)
    community_views.to_csv(
        args.output_filepath, sep="\t", float_format="%.4f", index=False
    )
    LOGGER.info("Done!")


if __name__ == "__main__":
    parser = configure_arguments()
    args = parser.parse_args()
    main(args)



"""Module for diversity's argument parser configuration.

Classes
-------
ValidateViewpoint
    Validator for -viewpoint parameter.

Functions
---------
configure_arguments
    Creates argument parser for .
"""

from argparse import Action, ArgumentParser
from sys import stdout
from warnings import warn

from numpy import inf

from greylock.exceptions import ArgumentWarning


class ValidateViewpoint(Action):
    """Validator for -viewpoint parameter."""

    def __call__(self, parser, args, values, option_string=None):
        """Validates -viewpoint parameter.

        Warns if arguments larger than 100 are passed, reminding that
        they are treated as infinity in the diversity calculation.
        """
        if any([viewpoint > 100 and viewpoint != inf for viewpoint in values]):
            warn(
                "viewpoints > 100.0 defaults to the analytical formula"
                " for viewpoint = infinity.",
                category=ArgumentWarning,
            )
        setattr(args, self.dest, values)


def configure_arguments():
    """Creates argument parser.

    Returns
    -------
    argparse.ArgumentParser configured to handle command-line arguments
    for executing diversity as a module.
    """
    parser = ArgumentParser()
    parser.add_argument(
        "-i",
        "--input_filepath",
        help=(
            "A csv or tsv file with one column per subcommunity, one "
            "row per species, where each element contains the count of "
            "each species in the corresponding subcommunities."
        ),
    )
    parser.add_argument(
        "-l",
        "--log_level",
        help=(
            "Logging verbosity level. Must be one of DEBUG, INFO,"
            " WARNING, ERROR, CRITICAL (listed in decreasing"
            " verbosity)."
        ),
        default="WARNING",
    )
    parser.add_argument(
        "-o",
        "--output_filepath",
        default=stdout,
        help="A filepath to where the program's output will be saved",
    )
    parser.add_argument(
        "-s",
        "--similarity",
        help=(
            "The filepath to a csv or tsv file containing a similarity"
            " for the species in the input file. The file must have a"
            " header listing the species names corresponding to each"
            " column, and column and row ordering must be the same."
        ),
    )
    parser.add_argument(
        "-v",
        "--viewpoint",
        nargs="+",
        type=float,
        help=(
            "A list of viewpoint parameters. Any non-negative number"
            " (including inf) is valid, but viewpoint parameters"
            " greater than 100 are treated like inf."
        ),
        action=ValidateViewpoint,
    )
    parser.add_argument(
        "-z",
        "--chunk_size",
        type=int,
        help="Number of rows to read at a time from the similarities matrix.",
        default=1,
    )
    return parser



from numpy import allclose
from pytest import mark

from greylock.metacommunity import Metacommunity
from greylock.components import Components
from greylock.abundance import make_abundance
from greylock.tests.base_tests.metacommunity_test import metacommunity_data
from greylock.tests.base_tests.similarity_test import similarity_array_3by3_1


@mark.parametrize(
    "data",
    metacommunity_data,
)
def test_make_components(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    assert isinstance(metacommunity.components, Components)


@mark.parametrize("data", metacommunity_data[2:])
def test_metacommunity_similarity(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    assert allclose(
        metacommunity.components.metacommunity_ordinariness,
        data.metacommunity_similarity,
    )


@mark.parametrize("data", metacommunity_data[2:])
def test_subcommunity_similarity(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    assert allclose(
        metacommunity.components.subcommunity_ordinariness, data.subcommunity_similarity
    )


@mark.parametrize("data", metacommunity_data[2:])
def test_normalized_subcommunity_similarity(data):
    metacommunity = Metacommunity(counts=data.counts, similarity=data.similarity)
    assert allclose(
        metacommunity.components.normalized_subcommunity_ordinariness,
        data.normalized_subcommunity_similarity,
    )



"""Custom exceptions raised by the diversity package.

Exceptions
----------
DiversityError
    Base class for all custom diversity exceptions.
DiversityWarning
    Base class for all custom diversity warnings.
InvalidArgumentError
    Raised when invalid argument is passed to a function.
ArgumentWarning
    Used for warnings of problematic argument choices.
"""


class DiversityError(Exception):
    pass


class DiversityWarning(Warning):
    pass


class InvalidArgumentError(DiversityError):
    pass


class ArgumentWarning(DiversityWarning):
    pass



"""Miscellaneous helper module for the metacommunity package.

Functions
---------
power_mean
    Calculates weighted power means.
"""

from numpy import (
    amax,
    amin,
    float64,
    inf,
    isclose,
    multiply,
    power,
    prod,
    sum as numpy_sum,
    zeros,
    all,
    any,
    ndarray,
)

from greylock.exceptions import InvalidArgumentError


def __validate_power_mean_args(
    weights: ndarray, items: ndarray, atol: float, weight_is_nonzero: ndarray
) -> None:
    """Validates arguments for power_mean.

    Parameters
    ----------
    weights, items, atol
        Same as for power_mean.
    weight_is_nonzero
        Boolean array of same shape as weights, indicating those whose
        absolute value meets or exceeds atol.

    Raises
    ------
    InvalidArgumentError when:
    - weights has more than 2 axes.
    - weights and items do not have the same shape
    - any column in weights contains 0s only when weights is 2d and when
      weights is all 0.
    """
    if len(weights.shape) > 2:
        raise InvalidArgumentError(
            f"'weights' shape must have 1 or 2 dimensions, but 'weights' had shape {weights.shape}."
        )
    if weights.shape != items.shape:
        raise InvalidArgumentError(
            f"Shape of 'weights' ({weights.shape}) must be the same as"
            f" shape of 'items' ({items.shape})."
        )
    all_0_column = all(~weight_is_nonzero, axis=0)
    if any(all_0_column):
        raise InvalidArgumentError(
            "Argument 'weights' must have at least one nonzero weight in each column. A weight is"
            " considered 0 if its absolute value is greater than or equal to"
            f" configurable minimum threshold: {atol:.2e}."
        )


def power_mean(
    order: float, weights: ndarray, items: ndarray, atol: float = 1e-9
) -> ndarray:
    """Calculates weighted power means.

    Parameters
    ----------
    order:
        Exponent used for the power mean.
    weights:
        The weights corresponding to items. Must be 1-d or 2-d. If 2-d,
        each column must contain at least one weight whose absolute
        value is greater than or equal to atol.
    items:
        The elements for which the weighted power mean is computed. Must
        have same shape as weights.
    atol:
        Threshold below which weights are considered to be 0.

    Returns
    -------
    A numpy.ndarray of the power means of items along axis 0 using order
    as exponent, weighted by weights. The array shape is the same as
    that of weights, or items except with the 0-axis removed. In the
    case of 1-d weights and items, the result has shape (1,). When order
    is close to 0 (absolute value less than atol), less than -100, or
    greater than 100 analytical formulas for the limits at 0, -infinity,
    or infinity are used respectively. An exception is raised if all weights
    in a column are close to 0.
    """
    weight_is_nonzero = abs(weights) >= atol
    __validate_power_mean_args(weights, items, atol, weight_is_nonzero)
    if isclose(order, 0):
        return prod(
            power(items, weights, where=weight_is_nonzero),
            axis=0,
            where=weight_is_nonzero,
        )
    elif order < -100:
        return amin(items, axis=0, where=weight_is_nonzero, initial=inf)
    elif order > 100:
        return amax(items, axis=0, where=weight_is_nonzero, initial=-inf)
    else:
        result = zeros(shape=items.shape, dtype=float64)
        power(items, order, where=weight_is_nonzero, out=result)
        multiply(result, weights, where=weight_is_nonzero, out=result)
        items_sum = numpy_sum(result, axis=0, where=weight_is_nonzero)
        return power(items_sum, 1 / order)



from typing import List, Any, Callable, Union
from numpy import ndarray, empty, concatenate, float64, vstack, zeros
from pandas import DataFrame
from greylock.exceptions import InvalidArgumentError

# avoid mypy error: see https://github.com/jorenham/scipy-stubs/issues/100
from scipy.sparse import spmatrix  # type: ignore
import ray  # type: ignore
from greylock.similarity import (
    SimilarityFromFunction,
    SimilarityFromSymmetricFunction,
    weighted_similarity_chunk_nonsymmetric,
    weighted_similarity_chunk_symmetric,
)


class SimilarityFromRayFunction(SimilarityFromFunction):
    """Implements Similarity by calculating similarities with a callable
    function."""

    def __init__(
        self,
        func: Callable,
        X: Union[ndarray, DataFrame],
        chunk_size: int = 100,
        max_inflight_tasks: int = 64,
        similarities_out: Union[ndarray, None] = None,
    ) -> None:
        """
        similarity:
            A Callable that calculates similarity between a pair of
            species. Must take two rows from X and return a numeric
            similarity value.
            If X is a 2D array, the rows will be 1D arrays.
            If X is a DataFrame, the rows will be named tuples.
        X:
            An array or DataFrame where each row contains the feature values
            for a given species.
        chunk_size:
            Determines how many rows of the similarity matrix each will
            be processes at a time. In general, choosing a larger
            chunk_size will make the calculation faster, but will also
            require more memory.
        """
        super().__init__(func, X, chunk_size, similarities_out)
        self.max_inflight_tasks = max_inflight_tasks

    def get_Y(self):
        return None

    def weighted_abundances(
        self,
        relative_abundance: Union[ndarray, spmatrix],
    ) -> ndarray:
        weighted_similarity_chunk = ray.remote(weighted_similarity_chunk_nonsymmetric)
        X_ref = ray.put(self.X)
        Y_ref = ray.put(self.get_Y())
        abundance_ref = ray.put(relative_abundance)
        futures: List[Any] = []
        results = []

        def process_refs(refs):
            nonlocal results
            for chunk_index, abundance_chunk, similarity_chunk in ray.get(refs):
                results.append((chunk_index, abundance_chunk))
                if self.similarities_out is not None:
                    self.similarities_out[
                        chunk_index : chunk_index + similarity_chunk.shape[0], :
                    ] = similarity_chunk

        for chunk_index in range(0, self.X.shape[0], self.chunk_size):
            if len(futures) >= self.max_inflight_tasks:
                ready_refs, futures = ray.wait(futures)
                process_refs(ready_refs)
            chunk_future = weighted_similarity_chunk.remote(
                similarity=self.func,
                X=X_ref,
                Y=Y_ref,
                relative_abundance=abundance_ref,
                chunk_size=self.chunk_size,
                chunk_index=chunk_index,
                return_Z=(self.similarities_out is not None),
            )
            futures.append(chunk_future)
        process_refs(futures)
        results.sort()  # This sorts by chunk index (1st in tuple)
        weighted_similarity_chunks = [r[1] for r in results]
        return concatenate(weighted_similarity_chunks)


class IntersetSimilarityFromRayFunction(SimilarityFromRayFunction):
    def __init__(
        self,
        func: Callable,
        X: Union[ndarray, DataFrame],
        Y: Union[ndarray, DataFrame],
        chunk_size: int = 100,
        max_inflight_tasks=64,
        similarities_out: Union[ndarray, None] = None,
    ):
        super().__init__(func, X, chunk_size, max_inflight_tasks, similarities_out)
        self.Y = Y

    def get_Y(self):
        return self.Y

    def self_similar_weighted_abundances(
        self, relative_abundance: Union[ndarray, spmatrix]
    ):
        raise InvalidArgumentError(
            "Inappropriate similarity class for diversity measures"
        )


class SimilarityFromSymmetricRayFunction(SimilarityFromSymmetricFunction):
    """Implements Similarity by calculating similarities with a callable
    function for one triangle of the similarity matrix, and re-using those
    values for the other triangle.
    """

    def __init__(
        self,
        func: Callable,
        X: Union[ndarray, DataFrame],
        chunk_size: int = 100,
        max_inflight_tasks: int = 64,
        similarities_out: Union[ndarray, None] = None,
    ) -> None:
        super().__init__(func, X, chunk_size, similarities_out)
        self.max_inflight_tasks = max_inflight_tasks

    def weighted_abundances(
        self,
        relative_abundance: Union[ndarray, spmatrix],
    ) -> ndarray:
        weighted_similarity_chunk = ray.remote(weighted_similarity_chunk_symmetric)
        X_ref = ray.put(self.X)
        abundance_ref = ray.put(relative_abundance)
        futures: List[Any] = []
        result = relative_abundance
        if self.similarities_out is not None:
            self.similarities_out.fill(0.0)

        def process_refs(refs):
            nonlocal result
            for chunk_index, addend, similarities in ray.get(refs):
                result = result + addend
                if self.similarities_out is not None:
                    self.similarities_out[
                        chunk_index : chunk_index + similarities.shape[0], :
                    ] = similarities

        for chunk_index in range(0, self.X.shape[0], self.chunk_size):
            if len(futures) >= self.max_inflight_tasks:
                (ready_refs, futures) = ray.wait(futures)
                process_refs(ready_refs)

            chunk_future = weighted_similarity_chunk.remote(
                similarity=self.func,
                X=X_ref,
                relative_abundance=abundance_ref,
                chunk_size=self.chunk_size,
                chunk_index=chunk_index,
                return_Z=(self.similarities_out is not None),
            )
            futures.append(chunk_future)
        process_refs(futures)
        if self.similarities_out is not None:
            self.similarities_out += self.similarities_out.T
            for i in range(self.X.shape[0]):
                self.similarities_out[i, i] = 1.0
        return result
